from starbot.core import commands  # isort:skip
from starbot.core.bot import StarBot  # isort:skip
import discord  # isort:skip
import typing  # isort:skip

import asyncio
import datetime
import inspect
import logging
import os
import random
import re
import string
from copy import copy
from functools import partial
from pathlib import Path
from random import choice

import aiohttp
from starbot.core.utils.chat_formatting import humanize_list
from starbot.logging import RotatingFileHandler

from .views import ConfirmationAskView

__all__ = ["CogsUtils"]


def _(untranslated: str) -> str:
    return untranslated


replacement_var_paths: bool = True


class CogsUtils:
    """Utils for StarCogs!"""

    # def __init__(
    #     self, cog: typing.Optional[commands.Cog] = None, bot: typing.Optional[StarBot] = None
    # ) -> None:
    #     if cog is not None:
    #         if isinstance(cog, str):
    #             cog = bot.get_cog(cog)
    #         self.cog: commands.Cog = cog
    #         self.bot: StarBot = self.cog.bot if hasattr(self.cog, "bot") else bot

    #     elif bot is not None:
    #         self.cog: typing.Optional[commands.Cog] = None
    #         self.bot: StarBot = bot
    #     else:
    #         self.cog: typing.Optional[commands.Cog] = None
    #         self.bot: typing.Optional[StarBot] = None

    @property
    def is_dpy2(self) -> bool:
        """
        Returns `True` if the current starbot instance is running under dpy2.
        """
        return discord.version_info.major >= 2

    @classmethod
    def replace_var_paths(cls, text: str, reverse: typing.Optional[bool] = False) -> str:
        if not reverse:
            if not replacement_var_paths:
                return text
            for env_var in ("USERPROFILE", "HOME", "USERNAME", "COMPUTERNAME"):
                if env_var in os.environ:
                    regex = re.compile(re.escape(os.environ[env_var]), re.I)
                    text = regex.sub(f"{{{env_var}}}", text)
                    regex = re.compile(re.escape(os.environ[env_var].replace("\\", "\\\\")), re.I)
                    text = regex.sub(f"{{{env_var}}}", text)
                    regex = re.compile(re.escape(os.environ[env_var].replace("\\", "/")), re.I)
                    text = regex.sub(f"{{{env_var}}}", text)
        else:

            class FakeDict(typing.Dict):
                def __missing__(self, key: str) -> str:
                    if (
                        key.upper() in ("USERPROFILE", "HOME", "USERNAME", "COMPUTERNAME")
                        and key.upper() in os.environ
                    ):
                        return os.environ[key.upper()]
                    return f"{{{key}}}"

            text = text.format_map(FakeDict())
        return text

    # async def add_cog(
    #     self, bot: typing.Optional[StarBot] = None, cog: typing.Optional[commands.Cog] = None, **kwargs
    # ) -> commands.Cog:
    #     """
    #     Load a cog by checking whether the required function is awaitable or not.
    #     """
    #     if bot is None:
    #         bot = self.bot
    #     if cog is None:
    #         cog = self.cog
    #         await self.change_config_unique_identifier(cog=cog)
    #         self._setup()
    #     value = bot.add_cog(cog, **kwargs)
    #     if inspect.isawaitable(value):
    #         await value
    #     return cog

    @classmethod
    def get_logger(
        cls, name: typing.Optional[str] = None, cog: typing.Optional[commands.Cog] = None
    ) -> logging.Logger:
        """
        Get a logger for a provided name or a provided cog.
        Thanks to @laggron42 on GitHub! (https://github.com/laggron42/Laggron-utils/blob/master/laggron_utils/logging.py)
        """
        if name is None and cog is None:
            raise RuntimeError()
        if name is not None and cog is None:
            return (
                logging.getLogger(name)
                if name.startswith("star.")
                else logging.getLogger(f"star.StarCogs.{name}")
            )

        logger = (
            logging.getLogger(f"star.{cog.__repo_name__}.{cog.qualified_name}")
            if name is None
            else logging.getLogger(f"star.{cog.__repo_name__}.{cog.qualified_name}.{name}")
        )

        __log = partial(logging.Logger._log, logger)

        def _log(level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1):
            from logging import CRITICAL, DEBUG, ERROR, FATAL, INFO, WARN, WARNING

            VERBOSE = DEBUG - 3
            TRACE = DEBUG - 5
            levels = {
                CRITICAL: "CRITICAL",
                DEBUG: "DEBUG",
                ERROR: "ERROR",
                FATAL: "FATAL",
                INFO: "INFO",
                WARN: "WARN",
                WARNING: "WARNING",
                VERBOSE: "VERBOSE",
                TRACE: "TRACE",
            }
            _level = levels.get(level, str(level))
            if _level not in cog.logs:
                cog.logs[_level] = []
            cog.logs[_level].append(
                {
                    "time": datetime.datetime.now(),
                    "level": level,
                    "message": msg,
                    "args": args,
                    "exc_info": exc_info,
                    "levelname": _level,
                }
            )
            __log(
                level=level,
                msg=msg,
                args=args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
                stacklevel=stacklevel,
            )

        setattr(logger, "_log", _log)

        # Logging in a log file.
        # (File is automatically created by the module, if the parent foler exists.)
        try:
            formatter = logging.Formatter(
                "[{asctime}] {levelname} [{name}] {message}",
                datefmt="%Y-%m-%d %H:%M:%S",
                style="{",
            )
            if cog.data_path.exists():
                file_handler = RotatingFileHandler(
                    stem=cog.qualified_name,
                    directory=cog.data_path,
                    maxBytes=1_000_0,
                    backupCount=0,
                    encoding="utf-8",
                )
                # file_handler.doRollover()
                file_handler.setLevel(logging.DEBUG)
                file_handler.setFormatter(formatter)
                logger.addHandler(file_handler)
        except Exception as e:
            logger.debug("Error when initiating the logger in a separate file.", exc_info=e)

        return logger

    @classmethod
    def close_logger(cls, logger: logging.Logger) -> None:
        """
        Closes the files for the logger of a cog.
        """
        for handler in logger.handlers:
            handler.close()
        logger.handlers = []

    @classmethod
    async def get_cog_version(
        cls, bot: StarBot, cog: typing.Union[commands.Cog, str]
    ) -> typing.Tuple[int, float, str]:
        cog_name = cog.lower() if isinstance(cog, str) else cog.qualified_name.lower()
        downloader_cog = bot.get_cog("Downloader")
        if downloader_cog is None:
            raise RuntimeError("The Downloader cog is not loaded.")

        if await bot._cog_mgr.find_cog(cog_name) is None:
            raise ValueError("This cog was not found in any cog path.")

        from starbot.cogs.downloader.repo_manager import ProcessFormatter, Repo

        repo = None
        path = Path(inspect.getsourcefile(cog.__class__))
        if not path.parent.parent == (await bot._cog_mgr.install_path()):
            local = None
            repo = Repo(name="", url="", branch="", commit="", folder_path=path.parent.parent)
        else:
            local = discord.utils.get(await downloader_cog.installed_cogs(), name=cog_name)
            if local is not None:
                repo = local.repo
        if repo is None:
            raise ValueError("This cog is not installed on this bot with Downloader.")

        exists, __ = repo._existing_git_repo()
        if not exists:
            raise ValueError(f"A git repo does not exist at path: {repo.folder_path}")
        git_command = ProcessFormatter().format(
            "git -C {path} rev-list HEAD --count {cog_name}",
            path=repo.folder_path,
            cog_name=cog_name,
        )
        p = await repo._run(git_command)
        if p.returncode != 0:
            raise asyncio.IncompleteReadError(
                "No results could be retrieved from the git command.", None
            )
        nb_commits = p.stdout.decode(encoding="utf-8").strip()
        nb_commits = int(nb_commits)

        version = round(1.0 + (nb_commits / 100), 2)

        if local is not None:
            commit = local.commit
        else:
            git_command = ProcessFormatter().format(
                "git -C {path} log HEAD -1 {cog_name}", path=repo.folder_path, cog_name=cog_name
            )
            p = await repo._run(git_command)
            if p.returncode != 0:
                raise asyncio.IncompleteReadError(
                    "No results could be retrieved from the git command.", None
                )
            commit = p.stdout.decode(encoding="utf-8").strip()
            commit = commit.split("\n")[0][7:]

        return nb_commits, version, commit

    @classmethod
    async def check_if_to_update(
        cls,
        bot: StarBot,
        cog: typing.Union[commands.Cog, str],
        repo_url: typing.Optional[str] = None,
    ) -> typing.Tuple[bool, str, str]:
        cog_name = cog.lower() if isinstance(cog, str) else cog.qualified_name.lower()
        if repo_url is None:
            downloader_cog = bot.get_cog("Downloader")
            if downloader_cog is None:
                raise RuntimeError("The Downloader cog is not loaded.")
            if await bot._cog_mgr.find_cog(cog_name) is None:
                raise ValueError("This cog was not found in any cog path.")
            local = discord.utils.get(await downloader_cog.installed_cogs(), name=cog_name)
            if local is None:
                raise ValueError("This cog is not installed on this bot with the cog Downloader.")
            local_commit = local.commit
            repo = local.repo
            if repo is None:
                raise ValueError("This cog is not installed on this bot with the cog Downloader.")
            repo_url = repo.url
        else:
            cog = None
            cog_name = None

        if isinstance(repo_url, str):
            repo_owner, repo_name, repo_branch = (
                re.compile(
                    r"(?:https?:\/\/)?git(?:hub|lab).com\/(?P<repo_owner>[A-z0-9-_.]*)\/(?P<repo>[A-z0-9-_.]*)(?:\/tree\/(?P<repo_branch>[A-z0-9-_.]*))?",
                    re.I,
                ).findall(repo_url)
            )[0]
        else:
            repo_owner, repo_name, repo_branch = repo_url
            repo_branch = repo_branch or repo.branch
        async with aiohttp.ClientSession() as session:
            async with session.get(
                (
                    f"https://api.github.com/repos/{repo_owner}/{repo_name}/commits?sha={repo_branch}&path={cog_name}"  # Thanks Jack!
                    if repo_branch
                    else f"https://api.github.com/repos/{repo_owner}/{repo_name}/commits?path={cog_name}"
                ),  # f"https://api.github.com/repos/{repo_owner}/{repo_name}/git/refs/heads/{repo_branch}" & f"https://api.github.com/repos/{repo_owner}/{repo_name}/contents?path={cog_name}"
                timeout=3,
            ) as r:
                online = await r.json()
        if (
            isinstance(online, typing.Dict)
            and "message" in online
            and "API rate limit exceeded" in online["message"]
        ):
            raise asyncio.LimitOverrunError("API rate limit exceeded.", 47)
        if online is None or not isinstance(online, typing.List) or len(online) == 0:
            raise asyncio.IncompleteReadError(
                "No results could be retrieved from the git API.", None
            )
        online_commit = online[0]["sha"]

        async def compare_commit_dates(repo_owner, repo_name, commit_sha1, commit_sha2):
            async def get_commit_date(
                repo_owner: str, repo_name: str, commit_sha: str, session: aiohttp.ClientSession
            ):
                url = f"https://api.github.com/repos/{repo_owner}/{repo_name}/commits/{commit_sha}"
                headers = {"Accept": "application/vnd.github+json"}
                async with session.get(url, headers=headers) as response:
                    data = await response.json()
                    if "commit" not in data:
                        raise asyncio.TimeoutError(
                            "No results could be retrieved from the git API."
                        )
                    commit_date = data["commit"]["committer"]["date"]
                return commit_date

            async with aiohttp.ClientSession() as session:
                commit_date1 = await get_commit_date(repo_owner, repo_name, commit_sha1, session)
                commit_date2 = await get_commit_date(repo_owner, repo_name, commit_sha2, session)
                if commit_date1 > commit_date2:
                    # Commit `{commit_sha1}` is newer than commit `{commit_sha2}`.
                    return False
                elif commit_date1 < commit_date2:
                    # Commit `{commit_sha2}` is newer than commit `{commit_sha1}`.
                    return True
                else:
                    # Commits `{commit_sha1}` and `{commit_sha2}`are the same date.
                    return None

        try:
            to_update = await compare_commit_dates(
                repo_owner=repo_owner,
                repo_name=repo_name,
                commit_sha1=local_commit,
                commit_sha2=online_commit,
            )
        except ValueError:  # Failed API request (temporary).
            to_update = False
        else:
            path = Path(inspect.getsourcefile(cog.__class__))
            if not path.parent.parent == (await bot._cog_mgr.install_path()):
                to_update = False

        return to_update, local_commit, online_commit  # , online_commit_for_each_files

    @classmethod
    async def add_hybrid_commands(cls, bot: StarBot, cog: commands.Cog) -> None:
        if hasattr(cog, "settings") and hasattr(cog.settings, "commands_added"):
            await cog.settings.commands_added.wait()
        if cog.qualified_name == "Medicat" and hasattr(cog, "CC_added"):
            await cog.CC_added.wait()
        for _object in cog.walk_commands():
            if isinstance(_object, (commands.HybridCommand, commands.HybridGroup)):
                if _object.app_command is not None:
                    _object.app_command.description = _object.app_command.description.split("\n")[
                        0
                    ][:100]
                if _object.parent is not None and not _object.parent.invoke_without_command:
                    _object.checks.extend(_object.parent.checks)
        await bot.tree.red_check_enabled()

    @classmethod
    async def ConfirmationAsk(
        cls,
        ctx: commands.Context,
        *args,
        timeout: typing.Optional[int] = 60,
        timeout_message: typing.Optional[str] = _("Timed out, please try again"),
        way: typing.Optional[typing.Literal["buttons", "message"]] = "buttons",  # , "reactions"
        delete_message: typing.Optional[bool] = True,
        members_authored: typing.Optional[typing.Iterable[discord.Member]] = [],
        **kwargs,
    ) -> bool:
        """
        Request a confirmation by the user., in the form of buttons/message, with many additional options.
        """
        check_owner = True

        if way == "buttons":
            return await ConfirmationAskView(
                ctx=ctx,
                timeout=timeout,
                timeout_message=timeout_message,
                delete_message=delete_message,
                members=members_authored,
            ).start(*args, **kwargs)

        # elif way == "reactions":
        #     message = await ctx.send(*args, **kwargs)
        #     try:
        #         start_adding_reactions(message, reactions)
        #     except discord.HTTPException:
        #         way = "message"
        #     view = Reactions(
        #         bot=ctx.bot,
        #         message=message,
        #         remove_reaction=False,
        #         timeout=timeout,
        #         reactions=reactions,
        #         members=[ctx.author.id] + list(ctx.bot.owner_ids)
        #         if check_owner
        #         else [] + [x.id for x in members_authored],
        #     )
        #     try:
        #         reaction, user, function_result = await view.wait_result()
        #         if str(reaction.emoji) == reactions[0]:
        #             end_reaction = True
        #             if delete_message:
        #                 await cls.delete_message(message)
        #             return True
        #         elif str(reaction.emoji) == reactions[1]:
        #             end_reaction = True
        #             if delete_message:
        #                 await cls.delete_message(message)
        #             return False
        #     except TimeoutError:
        #         if delete_message:
        #             await cls.delete_message(message)
        #         if timeout_message is not None:
        #             await ctx.send(timeout_message)
        #         return None

        elif way == "message":
            message = await ctx.send(*args, **kwargs)

            def check(msg):
                if check_owner:
                    return (
                        msg.author.id == ctx.author.id
                        or msg.author.id in ctx.bot.owner_ids
                        or msg.author.id in [x.id for x in members_authored]
                        and msg.channel == ctx.channel
                        and msg.content in ("yes", "y", "no", "n")
                    )
                else:
                    return (
                        msg.author.id == ctx.author.id
                        or msg.author.id in [x.id for x in members_authored]
                        and msg.channel == ctx.channel
                        and msg.content in ("yes", "y", "no", "n")
                    )

            try:
                end_reaction = False
                msg = await ctx.bot.wait_for("message", timeout=timeout, check=check)
                if msg.content in ("yes", "y"):
                    end_reaction = True
                    if delete_message:
                        await cls.delete_message(message)
                    await cls.delete_message(msg)
                    return True
                elif msg.content in ("no", "n"):
                    end_reaction = True
                    if delete_message:
                        await cls.delete_message(message)
                    await cls.delete_message(msg)
                    return False
            except asyncio.TimeoutError:
                if not end_reaction:
                    if delete_message:
                        await cls.delete_message(message)
                    if timeout_message is not None:
                        await ctx.send(timeout_message)
                    return None

    @classmethod
    async def delete_message(
        cls, message: discord.Message, delay: typing.Optional[float] = None
    ) -> bool:
        """
        Delete a message, ignoring any exceptions.
        Easier than putting these 3 lines at each message deletion for each cog.
        """
        if message is None:
            return None
        try:
            await message.delete(delay=delay)
        except discord.NotFound:  # Already deleted.
            return True
        except discord.HTTPException:
            return False
        else:
            return True

    @classmethod
    async def invoke_command(
        cls,
        bot: StarBot,
        author: discord.User,
        channel: discord.TextChannel,
        command: str,
        prefix: typing.Optional[str] = None,
        message: typing.Optional[discord.Message] = None,
        dispatch_message: typing.Optional[bool] = False,
        invoke: typing.Optional[bool] = True,
        __is_mocked__: typing.Optional[bool] = True,
        created_at: typing.Optional[datetime.datetime] = None,
        **kwargs,
    ) -> typing.Union[commands.Context, discord.Message]:
        """
        Invoke the specified command with the specified user in the specified channel.
        """
        if created_at is None:
            created_at = datetime.datetime.now(tz=datetime.timezone.utc)
        message_id = discord.utils.time_snowflake(created_at)
        if prefix == "/":  # For hybrid and slash commands.
            prefix = None
        if prefix is None:
            prefixes = await bot.get_valid_prefixes(guild=channel.guild)
            prefix = prefixes[0] if len(prefixes) < 3 else prefixes[2]
        old_content = f"{command}"
        content = f"{prefix}{old_content}"

        if message is None:
            message_content = content
            author_dict = {
                "id": f"{author.id}",
                "username": author.display_name,
                "avatar": author.avatar,
                "avatar_decoration": None,
                "discriminator": f"{author.discriminator}",
                "public_flags": author.public_flags,
                "bot": author.bot,
            }
            channel_id = channel.id
            timestamp = str(created_at).replace(" ", "T") + "+00:00"
            data = {
                "id": message_id,
                "type": 0,
                "content": message_content,
                "channel_id": f"{channel_id}",
                "author": author_dict,
                "attachments": [],
                "embeds": [],
                "mentions": [],
                "mention_roles": [],
                "pinned": False,
                "mention_everyone": False,
                "tts": False,
                "timestamp": timestamp,
                "edited_timestamp": None,
                "flags": 0,
                "components": [],
                "referenced_message": None,
            }
            message: discord.Message = discord.Message(
                channel=channel, state=bot._connection, data=data
            )
        else:
            message = copy(message)
            message.author = author
            message.channel = channel
            message.content = content

        context: commands.Context = await bot.get_context(message)
        context.author = author
        context.guild = channel.guild
        context.channel = channel

        if (  # StarBot's Alias
            not context.valid
            and context.prefix is not None
            and (alias_cog := bot.get_cog("Alias")) is not None
            and not await bot.cog_disabled_in_guild(alias_cog, context.guild)
        ):
            alias = await alias_cog._aliases.get_alias(context.guild, context.invoked_with)
            if alias is not None:

                async def command_callback(__, ctx: commands.Context):
                    await alias_cog.call_alias(ctx.message, ctx.prefix, alias)

                context.command = commands.command(name=command)(command_callback)
                context.command.cog = alias_cog
                context.command.params.clear()
                context.command.requires.ready_event.set()
        if (  # StarBot's CustomCommands
            not context.valid
            and context.prefix is not None
            and (custom_commands_cog := bot.get_cog("CustomCommands")) is not None
            and not await bot.cog_disabled_in_guild(custom_commands_cog, context.guild)
        ):
            try:
                raw_response, cooldowns = await custom_commands_cog.commandobj.get(
                    message=message, command=context.invoked_with
                )
                if isinstance(raw_response, list):
                    raw_response = random.choice(raw_response)
                elif isinstance(raw_response, str):
                    pass
            except Exception:
                pass
            else:

                async def command_callback(__, ctx: commands.Context):
                    # await custom_commands_cog.cc_callback(ctx)  # fake callback
                    try:
                        if cooldowns:
                            custom_commands_cog.test_cooldowns(
                                context, context.invoked_with, cooldowns
                            )
                    except Exception:
                        return
                    del ctx.args[0]
                    await custom_commands_cog.cc_command(
                        *ctx.args, **ctx.kwargs, raw_response=raw_response
                    )

                context.command = commands.command(name=command)(command_callback)
                context.command.cog = custom_commands_cog
                context.command.requires.ready_event.set()
                context.command.params = custom_commands_cog.prepare_args(raw_response)
        if (  # Phen/Lemon's Tags
            not context.valid
            and context.prefix is not None
            and (tags_cog := bot.get_cog("Tags")) is not None
            and not await bot.cog_disabled_in_guild(tags_cog, context.guild)
        ):
            tag = tags_cog.get_tag(context.guild, context.invoked_with, check_global=True)
            if tag is not None:
                message.content = f"{context.prefix}invoketag {command}"
                context: commands.Context = await bot.get_context(message)
                context.author = author
                context.guild = channel.guild
                context.channel = channel

        if __is_mocked__:
            context.__is_mocked__ = True
        context.__dict__.update(**kwargs)
        if not invoke:
            return context
        if context.valid:
            await bot.invoke(context)
        else:
            if dispatch_message:
                message.content = old_content
                bot.dispatch("message", message)
        return context

    @classmethod
    async def get_hook(cls, bot: StarBot, channel: discord.TextChannel) -> discord.Webhook:
        """
        Create a discord.Webhook object. It tries to retrieve an existing webhook created by the bot or to create it itself.
        """
        hook = next(
            (webhook for webhook in await channel.webhooks() if webhook.user.id == bot.user.id),
            None,
        )
        if hook is None:
            hook = await channel.create_webhook(name=f"red_bot_hook_{str(channel.id)}")
        return hook

    @classmethod
    def get_embed(
        cls, embed_dict: typing.Dict
    ) -> typing.Dict[str, typing.Union[discord.Embed, str]]:
        data = embed_dict
        if data.get("embed"):
            data = data["embed"]
        elif data.get("embeds"):
            data = data.get("embeds")[0]
        if timestamp := data.get("timestamp"):
            data["timestamp"] = timestamp.strip("Z")
        if data.get("content"):
            content = data["content"]
            del data["content"]
        else:
            content = ""
        for x in data:
            if data[x] is None:
                del data[x]
            elif isinstance(data[x], typing.Dict):
                for y in data[x]:
                    if data[x][y] is None:
                        del data[x][y]
        try:
            embed = discord.Embed.from_dict(data)
            length = len(embed)
            if length > 6000:
                raise commands.BadArgument(
                    f"Embed size exceeds Discord limit of 6000 characters ({length})."
                )
        except Exception as e:
            raise commands.BadArgument(f"An error has occurred.\n{e}).")
        return {"embed": embed, "content": content}

    @classmethod
    def datetime_to_timestamp(
        cls,
        dt: datetime.datetime,
        format: typing.Literal["f", "F", "d", "D", "t", "T", "R"] = "f",
    ) -> str:
        """
        Generate a Discord timestamp from a datetime object.
        <t:TIMESTAMP:FORMAT>
        Parameters
        ----------
        dt : datetime.datetime
            The datetime object to use
        format : TimestampFormat, by default `f`
            The format to pass to Discord.
            - `f` short date time | `18 June 2021 02:50`
            - `F` long date time  | `Friday, 18 June 2021 02:50`
            - `d` short date      | `18/06/2021`
            - `D` long date       | `18 June 2021`
            - `t` short time      | `02:50`
            - `T` long time       | `02:50:15`
            - `R` relative time   | `8 days ago`
        Returns
        -------
        str
            Formatted timestamp
        Thanks to vexutils from Vexed01 in GitHub! (https://github.com/Vexed01/Vex-Cogs/blob/master/timechannel/vexutils/chat.py)
        """
        t = str(int(dt.timestamp()))
        return f"<t:{t}:{format}>"

    @classmethod
    def get_interval_string(
        cls,
        expires: typing.Optional[
            typing.Union[
                datetime.datetime, datetime.timedelta  # , dateutil.relativedelta.relativedelta
            ]
        ],
        utc_now: datetime.datetime = None,
        use_timestamp: bool = False,
    ) -> str:
        """
        Get a string from a given duration.
        """
        if expires is None:
            return "No future occurrence."
        if use_timestamp:
            expires = expires.replace(tzinfo=datetime.timezone.utc)
            return f"<t:{int(expires.timestamp())}:R>"
        if utc_now is None:
            utc_now = datetime.datetime.now(datetime.timezone.utc)
        if isinstance(expires, datetime.datetime):
            delta = utc_now - expires
            # delta.seconds = 0
        elif isinstance(expires, datetime.timedelta):
            delta = expires
        else:
            try:
                import dateutil
            except ImportError:
                delta = datetime.timedelta(seconds=expires)
            else:
                if isinstance(expires, dateutil.relativedelta.relativedelta):
                    delta = (utc_now + expires) - utc_now
                else:
                    delta = datetime.timedelta(seconds=expires)
        result = []
        total_secs = int(max(0, delta.total_seconds()))
        years, rem = divmod(total_secs, 3600 * 24 * 365)
        if years > 0:
            result.append(f"{years} year" + ("s" if years > 1 else ""))
        months, rem = divmod(rem, 3600 * 24 * 7 * 4)
        if months > 0:
            result.append(f"{months} month" + ("s" if months > 1 else ""))
        weeks, rem = divmod(rem, 3600 * 24 * 7)
        if weeks > 0:
            result.append(f"{weeks} week" + ("s" if weeks > 1 else ""))
        days, rem = divmod(rem, 3600 * 24)
        if days > 0:
            result.append(f"{days} day" + ("s" if days > 1 else ""))
        hours, rem = divmod(rem, 3600)
        if hours > 0:
            result.append(f"{hours} hour" + ("s" if hours > 1 else ""))
        minutes, rem = divmod(rem, 60)
        if minutes > 0:
            result.append(f"{minutes} minute" + ("s" if minutes > 1 else ""))
        seconds = rem
        if seconds > 0:
            result.append(f"{seconds} second" + ("s" if seconds > 1 else ""))
        return humanize_list(result) if result else "just now"  # "0 minute"

    @classmethod
    def check_permissions_for(
        cls,
        channel: typing.Union[discord.TextChannel, discord.VoiceChannel, discord.DMChannel],
        user: discord.User,
        check: typing.Union[typing.List, typing.Dict],
    ) -> bool:
        """
        Check all permissions specified as an argument.
        """
        if getattr(channel, "guild", None) is None:
            return True
        permissions = channel.permissions_for(user)
        if isinstance(check, typing.List):
            new_check = {p: True for p in check}
            check = new_check
        return not any(
            getattr(permissions, p, None)
            is not None  # Explicitly check whether the value is None.
            and (
                (check[p] and not getattr(permissions, p))
                or (not check[p] and getattr(permissions, p))
            )
            for p in check
        )

    # @classmethod
    # def create_loop(
    #     cls,
    #     cog: commands.Cog,
    #     function,
    #     name: typing.Optional[str] = None,
    #     days: typing.Optional[int] = 0,
    #     hours: typing.Optional[int] = 0,
    #     minutes: typing.Optional[int] = 0,
    #     seconds: typing.Optional[int] = 0,
    #     function_kwargs: typing.Optional[typing.Dict] = None,
    #     wait_raw: typing.Optional[bool] = False,
    #     limit_count: typing.Optional[int] = None,
    #     limit_date: typing.Optional[datetime.datetime] = None,
    #     limit_exception: typing.Optional[int] = None,
    # ) -> Loop:
    #     """
    #     Create a `Loop` class instance, with default values and loop object recording functionality.
    #     """
    #     if function_kwargs is None:
    #         function_kwargs = {}
    #     if name is None:
    #         name = cog.qualified_name
    #     if (
    #         datetime.timedelta(
    #             days=days, hours=hours, minutes=minutes, seconds=seconds
    #         ).total_seconds()
    #         == 0
    #     ):
    #         seconds = 900  # 15 minutes
    #     loop = Loop(
    #         cog=cog,
    #         name=name,
    #         function=function,
    #         days=days,
    #         hours=hours,
    #         minutes=minutes,
    #         seconds=seconds,
    #         function_kwargs=function_kwargs,
    #         wait_raw=wait_raw,
    #         limit_count=limit_count,
    #         limit_date=limit_date,
    #         limit_exception=limit_exception,
    #     )
    #     # if (existing_loop := discord.utils.get(self.cog.loops, name=name)) is not None:
    #     #     existing_loop.stop_all()
    #     cog.loops.append(loop)
    #     return loop

    @classmethod
    def get_all_repo_cogs_objects(cls, bot: StarBot) -> typing.Dict[str, commands.Cog]:
        """
        Get a dictionary containing the objects of all my cogs.
        """
        cogs = {}
        for cog in bot.cogs.values():
            if cog.qualified_name in ("CogGuide"):
                continue
            if getattr(cog, "__repo_name__", None) == "StarCogs":
                cogs[cog.qualified_name] = cog
        return cogs

    @classmethod
    def at_least_one_cog_loaded(cls, bot: StarBot) -> bool:
        """
        Return True if at least one cog of all my cogs is loaded.
        """
        return any(cog is not None for cog in cls.get_all_repo_cogs_objects(bot).values())

    @classmethod
    def generate_key(
        cls,
        length: typing.Optional[int] = 10,
        existing_keys: typing.Optional[typing.Union[typing.List, typing.Set]] = None,
        strings_used: typing.Optional[typing.List] = None,
    ) -> str:
        """
        Generate a secret key, with the choice of characters, the number of characters and a list of existing keys.
        """
        if existing_keys is None:
            existing_keys = []
        if strings_used is None:
            strings_used = {
                "ascii_lowercase": True,
                "ascii_uppercase": False,
                "digits": True,
                "punctuation": False,
                "others": [],
            }
        strings = []
        if "ascii_lowercase" in strings_used and strings_used["ascii_lowercase"]:
            strings += string.ascii_lowercase
        if "ascii_uppercase" in strings_used and strings_used["ascii_uppercase"]:
            strings += string.ascii_uppercase
        if "digits" in strings_used and strings_used["digits"]:
            strings += string.digits
        if "punctuation" in strings_used and strings_used["punctuation"]:
            strings += string.punctuation
        if "others" in strings_used and isinstance(strings_used["others"], typing.List):
            strings += strings_used["others"]
        while True:
            # This probably won't turn into an endless loop.
            key = "".join(choice(strings) for _ in range(length))
            if key not in existing_keys:
                return key

    @classmethod
    async def check_in_listener(
        cls,
        bot: StarBot,
        arg: typing.Union[discord.Message, discord.RawReactionActionEvent, discord.Interaction],
        allowed_by_whitelist_blacklist: bool = True,
    ) -> bool:
        """
        Check all parameters for the output of any listener.
        Thanks to Jack! (https://discord.com/channels/133049272517001216/160386989819035648/825373605000511518)
        """
        try:
            if isinstance(arg, discord.Message):
                # check whether the message was sent by a webhook
                if arg.webhook_id is not None:
                    raise discord.ext.commands.BadArgument()
                # check whether the message was sent in a guild
                if arg.guild is None:
                    raise discord.ext.commands.BadArgument()
                # check whether the message author isn't a bot
                if arg.author is None:
                    raise discord.ext.commands.BadArgument()
                if arg.author.bot:
                    raise discord.ext.commands.BadArgument()
                # check whether the bot can send messages in the given channel
                if arg.channel is None:
                    raise discord.ext.commands.BadArgument()
                if not arg.channel.permissions_for(arg.guild.me).send_messages:
                    raise discord.ext.commands.BadArgument()
                # check whether the cog isn't disabled
                # if self.cog is not None and await self.bot.cog_disabled_in_guild(
                #     self.cog, output.guild
                # ):
                #     raise discord.ext.commands.BadArgument()
                # check whether the channel isn't on the ignore list
                if not await bot.ignored_channel_or_guild(arg):
                    raise discord.ext.commands.BadArgument()
                # check whether the message author isn't on allowlist/blocklist
                if (
                    allowed_by_whitelist_blacklist
                    and not await bot.allowed_by_whitelist_blacklist(arg.author)
                ):
                    raise discord.ext.commands.BadArgument()
            elif isinstance(arg, discord.RawReactionActionEvent):
                # check whether the message was sent in a guild
                arg.guild = bot.get_guild(arg.guild_id)
                if arg.guild is None:
                    raise discord.ext.commands.BadArgument()
                # check whether the message author isn't a bot
                arg.author = arg.guild.get_member(arg.user_id)
                if arg.author is None:
                    raise discord.ext.commands.BadArgument()
                if arg.author.bot:
                    raise discord.ext.commands.BadArgument()
                # check whether the bot can send message in the given channel
                arg.channel = arg.guild.get_channel(arg.channel_id)
                if arg.channel is None:
                    raise discord.ext.commands.BadArgument()
                if not arg.channel.permissions_for(arg.guild.me).send_messages:
                    raise discord.ext.commands.BadArgument()
                # check whether the cog isn't disabled
                # if self.cog is not None and await self.bot.cog_disabled_in_guild(
                #     self.cog, output.guild
                # ):
                #     raise discord.ext.commands.BadArgument()
                # check whether the channel isn't on the ignore list
                if not await bot.ignored_channel_or_guild(arg):
                    raise discord.ext.commands.BadArgument()
                # check whether the message author isn't on allowlist/blocklist
                if (
                    allowed_by_whitelist_blacklist
                    and not await bot.allowed_by_whitelist_blacklist(arg.author)
                ):
                    raise discord.ext.commands.BadArgument()
            elif isinstance(arg, discord.Interaction):
                # check whether the message was sent in a guild
                if arg.guild is None:
                    raise discord.ext.commands.BadArgument()
                # check whether the message author isn't a bot
                if arg.author is None:
                    raise discord.ext.commands.BadArgument()
                if arg.author.bot:
                    raise discord.ext.commands.BadArgument()
                # check whether the bot can send message in the given channel
                if arg.channel is None:
                    raise discord.ext.commands.BadArgument()
                if not arg.channel.permissions_for(arg.guild.me).send_messages:
                    raise discord.ext.commands.BadArgument()
                # check whether the cog isn't disabled
                # if self.cog is not None and await self.bot.cog_disabled_in_guild(
                #     self.cog, output.guild
                # ):
                #     raise discord.ext.commands.BadArgument()
                # check whether the message author isn't on allowlist/blocklist
                if (
                    allowed_by_whitelist_blacklist
                    and not await bot.allowed_by_whitelist_blacklist(arg.author)
                ):
                    raise discord.ext.commands.BadArgument()
        except commands.BadArgument:
            return False
        return True



DOCS_BASE = "This cog has docs! Check them out at\nhttps://s.vexcodes.com/c/{}"

CHECK = "\N{HEAVY CHECK MARK}\N{VARIATION SELECTOR-16}"
CROSS = "\N{CROSS MARK}"

RED_CIRCLE = "\N{LARGE RED CIRCLE}"
GREEN_CIRCLE = "\N{LARGE GREEN CIRCLE}"

SNOWFLAKE_REGEX = r"\b\d{17,20}\b"



from starbot.core import commands  # isort:skip
from starbot.core.bot import StarBot  # isort:skip
from starbot.core import Config  # isort:skip
import typing  # isort:skip
import discord  # isort:skip

import asyncio
import re
from uuid import uuid4

import sentry_sdk
from starbot.core import __version__ as red_version
from starbot.core.utils.common_filters import INVITE_URL_RE

from .cogsutils import CogsUtils
from .loop import Loop

SNOWFLAKE_REGEX = r"\b\d{17,20}\b"
IP_V4_REGEX = (
    r"(\b25[0-5]|\b2[0-4][0-9]|\b[01]?[0-9][0-9]?)(\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)){3}"
)
IP_V6_REGEX = r"(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0-1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0-1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0-1}[0-9]){0,1}[0-9]))"

SENTRY_DSN = (
    "https://463fef380c8d6f900e3deec882110429@o4507885141753856.ingest.us.sentry.io/4507887488598016"
)

def _(untranslated: str) -> str:
    return untranslated

SENTRY_MASTER_MSG = _(
    "Hey there! This looks like the first time you're using Star's cogs (or you just updated to a"
    " version which supports this). To help make this cog, and all my others, as good and bug-free"
    " as possible, I have **opt-in** telemetry and error reporting __which affects all of my"
    " (github.com/LeDeathAmongst's) cogs__ on the Star-Cogs repository, using Sentry. The telemetry"
    " consists of data on the cog release and performance data of background tasks and loops (if"
    " applicable), and error reporting means that if something goes wrong the error and some"
    " associated data will be automatically sent to me so I can fix it quickly.\n\nA best effort"
    " is made to ensure no sensitive data is transmitted. For more information, including some"
    " technical details, visit"
    " <https://Star-cogs.readthedocs.io/en/latest/repo_telemetry.html>\n\n**If you would like to"
    " opt-in to telemetry and error reporting, and help me develop my cogs, run the command"
    " `[p]StarBot_Utils telemetrywithsentry True`. `[p]` is your prefix.**\nNo data is collected"
    " relating to command usage."
)
SENTRY_REMINDER_ON = _(
    "Hey there! You just installed Star's {} cog. This is a reminder that you previously enabled"
    " telemetry and error reporting, which applies to all of my cogs, and this one is no"
    " different.\n\nI would like to emphasize again that a best effort it made to remove sensitive"
    " data. You can see <https://Star-cogs.readthedocs.io/en/latest/repo_telemetry.html> for more"
    " details and change your choice at any time with the `[p]StarBot_Utils telemetrywithsentry"
    " False` command, applying to all my cogs."
)
SENTRY_REMINDER_OFF = _(
    "Hey there! You just installed Star's {} cog. This is a reminder that you previously chose"
    " not to enable telemetry and error reporting, which is also available in this cog. I hope you"
    " don't mind this reminder.\n\nI would like to emphasize again that a best effort it made to"
    " remove sensitive data. You can see"
    " <https://Star-cogs.readthedocs.io/en/latest/repo_telemetry.html> for more details and"
    " change your choice at any time with the `[p]StarBot_Utils telemetrywithsentry True` command,"
    " applying to all my cogs."
)

__all__ = ["SentryHelper"]

class SentryHelper:
    def __init__(self, bot: StarBot, cog: commands.Cog) -> None:
        if cog.qualified_name != "StarBot_Utils":
            raise ValueError(cog.qualified_name)
        self.bot: StarBot = bot
        self.cog: commands.Cog = cog

        self.last_errors: typing.Dict[
            str, typing.Dict[str, typing.Union[commands.Context, Exception]]
        ] = {}

        self.sentry_enabled: bool = None
        self.display_sentry_manual_command: bool = None
        self.send_reminders: bool = True
        self.uuid: typing.Optional[str] = None
        self.hubs: typing.Dict[str, sentry_sdk.Hub] = {}

        self.config: Config = cog.config
        self.sentry_global: typing.Dict[
            str, typing.Dict[str, typing.Union[int, bool, typing.Optional[str], typing.List[str]]]
        ] = {
            "sentry": {
                "version": 1,
                "sentry_enabled": False,
                "display_sentry_manual_command": True,
                "master_msg_sent": False,
                "uuid": None,
                "cogs_notified": [],
            }
        }
        self.config.register_global(**self.sentry_global)

        asyncio.create_task(self._async_init())
        self.cog.loops.append(
            Loop(
                cog=self.cog,
                name="Sentry Helper",
                function=self.periodic_session_restart,
                hours=1,
            )
        )

        self.dont_send_reminders: bool = False
        self.ready: asyncio.Event = asyncio.Event()

    async def _async_init(self) -> None:
        self.sentry_enabled = await self.config.sentry.sentry_enabled()
        self.display_sentry_manual_command = not self.sentry_enabled and (
            await self.config.sentry.display_sentry_manual_command()
        )
        uuid = await self.config.sentry.uuid()
        if uuid is None:
            uuid = str(uuid4())
            await self.config.sentry.uuid.set(uuid)
        self.uuid = uuid
        self.ready.set()

    async def periodic_session_restart(self) -> None:
        if not self.sentry_enabled:
            return
        for hub in self.hubs.values():
            hub.end_session()
            hub.start_session()

    async def send_command_error(
        self,
        ctx: commands.Context,
        error: commands.CommandError,
        manually: typing.Optional[bool] = False,
    ) -> typing.Optional[typing.Union[str, bool]]:
        try:
            if ctx.cog is None:
                return
            if not manually and not self.sentry_enabled:
                return False
            hub = await self.get_sentry_hub(ctx.cog)
            if hub is None:
                return
            if isinstance(error, commands.CommandInvokeError):
                if isinstance(ctx.command, discord.ext.commands.HybridCommand):
                    _type = "[hybrid|text]"
                else:
                    _type = "[text]"
            elif isinstance(error, commands.HybridCommandError):
                _type = "[hybrid|slash]"
            else:
                return False
            message = f"Error in {_type} command '{ctx.command.qualified_name}'."
            with sentry_sdk.isolation_scope() as scope:
                hub.add_breadcrumb(category="command", message=message)
                scope.set_extra("user_permissions", ctx.permissions)
                scope.set_extra("user_permissions_dict", dict(ctx.permissions))
                scope.set_extra("bot_permissions", ctx.bot_permissions)
                scope.set_extra("bot_permissions_dict", dict(ctx.bot_permissions))
                try:
                    e = error.original  # type:ignore
                except AttributeError:
                    e = error
                event_id = hub.capture_exception(e)
                scope._extras.clear()
            return event_id
        except Exception as e:
            if manually:
                raise e
            self.cog.logger.error("Sending an error to Sentry failed.", exc_info=e)
            return False

    def remove_sensitive_data(self, event: dict, hint: typing.Optional[dict] = {}) -> typing.Dict:
        """Remove sensitive data from the event. This should only be used by the Sentry SDK.
        This has two main parts:
        1) Remove any mentions of the bot's token
        2) Replace all IDs with a 4 digit number which originates from the timestamp of the ID
        3) Remove discord invites
        4) Replace var paths (userprofile, home, computer name, user name)
        Parameters
        ----------
        event : dict
            Event data
        hint : dict
            Event hint
        Returns
        -------
        dict
            The event dict with above stated sensitive data removed.
        """

        def regex_stuff(s: str) -> str:
            """Shorten any Discord IDs/snowflakes (basically any number 17-20 characters) to 4 digits
            by locating the timestamp and getting the last 4 digits - the milliseconds and the last
            digit of the second countsecond.
            Parameters
            ----------
            s : str
                String to shorten IDs in
            Returns
            -------
            str
                String with IDs shortened
            """
            s = re.sub(
                SNOWFLAKE_REGEX,
                lambda m: f"[SHORTENED-ID-{str(int(m.group()) >> 22)[-4:]}]",
                s,
            )
            s = re.sub(IP_V4_REGEX, "IP_V4", s)
            s = re.sub(IP_V6_REGEX, "IP_V6", s)
            return re.sub(INVITE_URL_RE, "[DISCORD-INVITE-LINK]", s)

        def recursive_replace(
            d: typing.Union[typing.Dict[str, typing.Any], typing.List, str], token: str
        ) -> typing.Union[dict, str]:
            """Recursively replace text in keys and values of a dictionary.
            Parameters
            ----------
            d : Union[Dict[str, Any], str]
                Dict or item to replace text in
            token : str
                Token to remove
            Returns
            -------
            dict
                Safe dict
            """
            if isinstance(d, typing.Dict):
                return {
                    CogsUtils.replace_var_paths(regex_stuff(k.replace(token, "[BOT-TOKEN]")))
                    if isinstance(k, str)
                    else k: recursive_replace(v, token)
                    for k, v in d.items()
                }
            elif isinstance(d, typing.List):
                return [
                    CogsUtils.replace_var_paths(regex_stuff(recursive_replace(i, token)))
                    if isinstance(i, str)
                    else recursive_replace(i, token)
                    for i in d
                ]
            return (
                CogsUtils.replace_var_paths(regex_stuff(d.replace(token, "[BOT_TOKEN]")))
                if isinstance(d, str)
                else d
            )

        token = self.bot.http.token

        return recursive_replace(event, token)  # type:ignore

    async def enable_sentry(self) -> None:
        """Enable Sentry telemetry and error reporting."""
        await self.config.sentry.sentry_enabled.set(True)
        self.sentry_enabled = True
        self.dont_send_reminders = False
        for hub in self.hubs.values():
            hub.start_session()

    async def disable_sentry(self) -> None:
        """Disable Sentry telemetry and error reporting."""
        await self.config.sentry.sentry_enabled.set(False)
        self.sentry_enabled = False
        self.dont_send_reminders = True

        for hub in self.hubs.values():
            hub.end_session()

    async def get_sentry_hub(
        self, cog: commands.Cog, force: typing.Optional[bool] = False
    ) -> sentry_sdk.Hub:
        """Get a Sentry Hub and Client for a DSN. Each cog should have its own hub.
        Returns
        -------
        Hub
            A Sentry Hub with a Client
        """
        if not self.ready.is_set():
            await self.ready.wait()
        if cog.qualified_name in self.hubs and not force:
            return self.hubs[cog.qualified_name]
        if getattr(cog, "__version__", None) is None and getattr(cog, "__commit__", None) is None:
            try:
                nb_commits, version, commit = await CogsUtils.get_cog_version(
                    bot=self.bot, cog=self.cog
                )
                cog.__version__ = version
                cog.__commit__ = commit
            except Exception:
                pass
        client = sentry_sdk.Client(
            dsn=SENTRY_DSN,
            traces_sample_rate=0.005,
            before_send=self.remove_sensitive_data,
            before_breadcrumb=self.remove_sensitive_data,
            release=(
                f"Star-Cogs|{cog.qualified_name}@{getattr(cog, '__version__', 1.0)}|{getattr(cog, '__commit__', '')}"
            ),
            debug=False,
            max_breadcrumbs=25,
            server_name="[EXPUNGED]",
        )

        scope = sentry_sdk.Scope()
        scope.set_tag("cog_name", cog.qualified_name)
        scope.set_tag("cog_version", getattr(cog, "__version__", 1.0))
        scope.set_tag("cog_commit", getattr(cog, "__commit__", ""))
        scope.set_tag("red_release", red_version)
        scope.set_tag("dpy_release", discord.__version__)
        scope.set_user({"id": self.uuid})

        hub = sentry_sdk.Hub(client, scope)
        self.hubs[cog.qualified_name] = hub

        hub.start_session()
        return hub

    async def cog_unload(self, cog: commands.Cog) -> sentry_sdk.Hub:
        """Close the linked Sentry Hub.
        Returns
        -------
        Hub
            A Sentry Hub closed
        """
        hub = self.hubs.get(cog.qualified_name, None)
        if hub is None:
            return None
        hub.end_session()
        hub.client.close()
        del self.hubs[cog.qualified_name]
        return hub

    async def maybe_send_owners(self, cog: commands.Cog) -> None:
        if not self.ready.is_set():
            await self.ready.wait()
        if (
            self.cog is None
            or not hasattr(self.cog, "telemetrywithsentry")
            or getattr(self.cog.telemetrywithsentry, "__is_dev__", False)
        ):
            return  # Not send automatically errors to Sentry for the moment.
        if not await self.config.sentry.master_msg_sent():
            self.dont_send_reminders = True
            await self.config.sentry.master_msg_sent.set(True)
            try:
                await self.bot.send_to_owners(SENTRY_MASTER_MSG.format(cog.qualified_name))
            except Exception:
                pass
            async with self.config.sentry.cogs_notified() as c_n:
                c_n.append(cog.qualified_name)
            return

        if cog.qualified_name in await self.config.sentry.cogs_notified():
            return
        if self.dont_send_reminders:
            async with self.config.sentry.cogs_notified() as c_n:
                c_n.append(cog.qualified_name)
            return
        try:
            if self.sentry_enabled:
                await self.bot.send_to_owners(SENTRY_REMINDER_ON.format(cog.qualified_name))
            else:
                self.dont_send_reminders = True
                await self.bot.send_to_owners(SENTRY_REMINDER_OFF.format(cog.qualified_name))
        except Exception:
            pass
        async with self.config.sentry.cogs_notified() as c_n:
            c_n.append(cog.qualified_name)



from __future__ import annotations

import asyncio
import json
from logging import Logger, getLogger
from pathlib import Path
from typing import Literal, NamedTuple

import aiohttp
from starbot.core import VersionInfo, commands
from starbot.core import version_info as cur_red_version
from rich import box as rich_box
from rich.table import Table  # type:ignore

from .chat import no_colour_rich_markup
from .consts import DOCS_BASE, GREEN_CIRCLE, RED_CIRCLE
from .loop import Loop

log = getLogger("star.utils")

cog_ver_lock = asyncio.Lock()

def get_vex_logger(name: str) -> Logger:
    """Get a logger for the given name.

    Parameters
    ----------
    name : str
        The ``__name__`` of the file

    Returns
    -------
    Logger
        The logger
    """
    final_name = "star."
    split = name.split(".")
    if len(split) == 2 and split[0] == split[1]:  # for example make `cmdlog.cmdlog` just `cmdlog`
        final_name += split[0]
    else:  # otherwise use full path
        final_name += name

    return getLogger(final_name)


# TODO: stop using red internal util

async def format_info(
    ctx: commands.Context,
    qualified_name: str,
    cog_version: str,
    extras: dict[str, str | bool] = {},
    loops: list[Loop] = [],
) -> str:
    """Generate simple info text about the cog. **Not** currently for use outside my cogs.

    Parameters
    ----------
    ctx : commands.Context
        Context
    qualified_name : str
        The name you want to show, eg "BetterUptime"
    cog_version : str
        The version of the cog
    extras : Dict[str, Union[str, bool]], optional
        Dict which is foramtted as key: value\\n. Bools as a value will be replaced with
        check/cross emojis, by default {}
    loops : List[Loop], optional
        List of Loops you want to show, by default []

    Returns
    -------
    str
        Simple info text.
    """
    cog_name = qualified_name.lower()
    current = _get_current_vers(cog_version, qualified_name)
    try:
        latest = await _get_latest_vers(cog_name)

        cog_updated = current.cog >= latest.cog
        utils_updated = current.utils == latest.utils
        red_updated = current.red >= latest.red
    except Exception:  # anything and everything, eg aiohttp error or version parsing error
        log.warning("Unable to parse versions.", exc_info=True)
        cog_updated, utils_updated, red_updated = "Unknown", "Unknown", "Unknown"
        latest = UnknownVers()

    start = f"{qualified_name} by Scarlet.\n<https://github.com/Scarleted01/Scarlet-Cogs>\n\n"

    main_table = Table(
        "", "Current", "Latest", "Up to date?", title="Versions", box=rich_box.MINIMAL
    )

    main_table.add_row(
        "This Cog",
        str(current.cog),
        str(latest.cog),
        GREEN_CIRCLE if cog_updated else RED_CIRCLE,
    )
    main_table.add_row(
        "Bundled Utils",
        current.utils,
        latest.utils,
        GREEN_CIRCLE if utils_updated else RED_CIRCLE,
    )
    main_table.add_row(
        "StarBot",
        str(current.red),
        str(latest.red),
        GREEN_CIRCLE if red_updated else RED_CIRCLE,
    )

    update_msg = "\n"
    if not cog_updated:
        update_msg += f"To update this cog, use the `{ctx.clean_prefix}cog update` command.\n"
    if not utils_updated:
        update_msg += (
            f"To update the bundled utils, use the `{ctx.clean_prefix}cog update` command.\n"
        )
    if not red_updated:
        update_msg += "To update StarBot, contact Scarlet\n"

    extra_table = Table("Key", "Value", title="Extras", box=rich_box.MINIMAL)

    data = []
    if loops:
        for loop in loops:
            extra_table.add_row(loop.friendly_name, GREEN_CIRCLE if loop.integrity else RED_CIRCLE)

    if extras:
        if data:
            extra_table.add_row("", "")
        for key, value in extras.items():
            if isinstance(value, bool):
                str_value = GREEN_CIRCLE if value else RED_CIRCLE
            else:
                assert isinstance(value, str)
                str_value = value
            extra_table.add_row(key, str_value)

    boxed = no_colour_rich_markup(main_table)
    boxed += update_msg
    if loops or extras:
        boxed += no_colour_rich_markup(extra_table)

    return f"{start}{boxed}"


async def out_of_date_check(cogname: str, currentver: str) -> None:
    """Send a log at warning level if the cog is out of date."""
    try:
        async with cog_ver_lock:
            vers = await _get_latest_vers(cogname)
        if VersionInfo.from_str(currentver) < vers.cog:
            log.warning(
                f"Your {cogname} cog, from Scarlet, is out of date. You can update your cogs with the "
                "'cog update' command in Discord."
            )
        else:
            log.debug(f"{cogname} cog is up to date")
    except Exception as e:
        log.debug(
            f"Something went wrong checking if {cogname} cog is up to date. See below.", exc_info=e
        )
        # really doesn't matter if this fails so fine with debug level
        return


class Vers(NamedTuple):
    cogname: str
    cog: VersionInfo
    utils: str
    red: VersionInfo


class UnknownVers(NamedTuple):
    cogname: str = "Unknown"
    cog: VersionInfo | Literal["Unknown"] = "Unknown"
    utils: str = "Unknown"
    red: VersionInfo | Literal["Unknown"] = "Unknown"


async def _get_latest_vers(cogname: str) -> Vers:
    data: dict
    async with aiohttp.ClientSession() as session:
        async with session.get(f"https://api.vexcodes.com/v2/vers/{cogname}", timeout=3) as r:
            data = await r.json()
            latest_utils = data["utils"][:7]
            latest_cog = VersionInfo.from_str(data.get(cogname, "0.0.0"))
        async with session.get("https://pypi.org/pypi/StarBot-DiscordBot/json", timeout=3) as r:
            data = await r.json()
            latest_red = VersionInfo.from_str(data.get("info", {}).get("version", "0.0.0"))

    return Vers(cogname, latest_cog, latest_utils, latest_red)


def _get_current_vers(curr_cog_ver: str, qual_name: str) -> Vers:
    with open(Path(__file__).parent / "commit.json") as fp:
        data = json.load(fp)
        latest_utils = data.get("latest_commit", "Unknown")[:7]

    return Vers(
        qual_name,
        VersionInfo.from_str(curr_cog_ver),
        latest_utils,
        cur_red_version,
    )



import datetime
from io import StringIO
from typing import Any, Literal, Sequence, Union

from starbot.core.utils.chat_formatting import box, humanize_list, humanize_number, inline
from rich.console import Console

TimestampFormat = Literal["f", "F", "d", "D", "t", "T", "R"]


def no_colour_rich_markup(*objects: Any, lang: str = "") -> str:
    """
    Slimmed down version of rich_markup which ensure no colours (/ANSI) can exist
    https://github.com/Cog-Creators/Red-DiscordBot/pull/5538/files (Kowlin)
    """
    temp_console = Console(  # Prevent messing with STDOUT's console
        color_system=None,
        file=StringIO(),
        force_terminal=True,
        width=80,
    )
    temp_console.print(*objects)
    return box(temp_console.file.getvalue(), lang=lang)  # type: ignore


def _hum(num: Union[int, float], unit: str, ndigits: int) -> str:
    """Round a number, then humanize."""
    return humanize_number(round(num, ndigits)) + f" {unit}"


def humanize_bytes(bytes: Union[int, float], ndigits: int = 0) -> str:
    """Humanize a number of bytes, rounding to ndigits. Only supports up to GB.

    This assumes 1GB = 1000MB, 1MB = 1000KB, 1KB = 1000B"""
    if bytes > 10000000000:  # 10GB
        gb = bytes / 1000000000
        return _hum(gb, "GB", ndigits)
    if bytes > 10000000:  # 10MB
        mb = bytes / 1000000
        return _hum(mb, "MB", ndigits)
    if bytes > 10000:  # 10KB
        kb = bytes / 1000
        return _hum(kb, "KB", ndigits)
    return _hum(bytes, "B", 0)  # no point in rounding


# maybe think about adding to core
def inline_hum_list(items: Sequence[str], *, style: str = "standard") -> str:
    """Similar to core's humanize_list, but all items are in inline code blocks. **Can** be used
    outside my cogs.

    Strips leading and trailing whitespace.

    Does not support locale.

    Does support style (see core's docs for available styles)

    Parameters
    ----------
    items : Sequence[str]
        The items to humanize
    style : str, optional
        The style. See core's docs, by default "standard"

    Returns
    -------
    str
        Humanized inline list.
    """
    inline_list = [inline(i.strip()) for i in items]
    return humanize_list(inline_list, style=style)


def datetime_to_timestamp(dt: datetime.datetime, format: TimestampFormat = "f") -> str:
    """Generate a Discord timestamp from a datetime object.

    <t:TIMESTAMP:FORMAT>

    Parameters
    ----------
    dt : datetime.datetime
        The datetime object to use
    format : TimestampFormat, by default `f`
        The format to pass to Discord.
        - `f` short date time | `18 June 2021 02:50`
        - `F` long date time  | `Friday, 18 June 2021 02:50`
        - `d` short date      | `18/06/2021`
        - `D` long date       | `18 June 2021`
        - `t` short time      | `02:50`
        - `T` long time       | `02:50:15`
        - `R` relative time   | `8 days ago`

    Returns
    -------
    str
        Formatted timestamp
    """
    t = str(int(dt.timestamp()))
    return f"<t:{t}:{format}>"



from starbot.core import commands  # isort:skip
import discord  # isort:skip
import typing  # isort:skip

# import typing_extensions  # isort:skip

import asyncio
import datetime
import inspect
import math
import time
import traceback
from io import StringIO

from starbot.core.utils.chat_formatting import box, pagify
from rich.console import Console
from rich.table import Table

from .cogsutils import CogsUtils

__all__ = ["Loop"]


def _(untranslated: str) -> str:
    return untranslated


def no_colour_rich_markup(
    *objects: typing.Any, lang: str = "", no_box: typing.Optional[bool] = False
) -> str:
    """
    Slimmed down version of rich_markup which ensure no colours (/ANSI) can exist
    https://github.com/Cog-Creators/Red-DiscordBot/pull/5538/files (Kowlin)
    """
    temp_console = Console(  # Prevent messing with STDOUT's console
        color_system=None,
        file=StringIO(),
        force_terminal=True,
        width=80,
    )
    temp_console.print(*objects)
    if no_box:
        return temp_console.file.getvalue()
    return box(temp_console.file.getvalue(), lang=lang)  # type: ignore


class Loop:
    """
    Create a loop, with many features.
    Thanks to Vexed01 on GitHub! (https://github.com/Vexed01/Vex-Cogs/blob/master/timechannel/loop.py and https://github.com/Vexed01/vex-cog-utils/vexutils/loop.py)
    """

    def __init__(
        self,
        cog: commands.Cog,
        name: str,
        function: typing.Callable,
        days: typing.Optional[int] = 0,
        hours: typing.Optional[int] = 0,
        minutes: typing.Optional[int] = 0,
        seconds: typing.Optional[int] = 0,
        function_kwargs: typing.Optional[typing.Dict[str, typing.Any]] = None,
        wait_raw: typing.Optional[bool] = False,
        limit_count: typing.Optional[int] = None,
        limit_date: typing.Optional[datetime.datetime] = None,
        limit_exception: typing.Optional[int] = None,
        start_now: typing.Optional[bool] = True,
    ) -> None:
        if function_kwargs is None:
            function_kwargs = {}
        self.cog: commands.Cog = cog

        self.name: str = name
        self.function: typing.Callable = function
        self.function_kwargs: typing.Dict[str, typing.Any] = function_kwargs
        self.interval: float = datetime.timedelta(
            days=days, hours=hours, minutes=minutes, seconds=seconds
        ).total_seconds()
        self.wait_raw: bool = wait_raw
        self.limit_count: int = limit_count
        self.limit_date: datetime.datetime = limit_date
        self.limit_exception: int = limit_exception
        self.stop_manually: bool = False

        self.start_datetime: datetime.datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        self.expected_interval = datetime.timedelta(seconds=self.interval)
        self.last_iteration: typing.Optional[datetime.datetime] = None
        self.next_iteration: typing.Optional[datetime.datetime] = None
        self.currently_running: bool = False  # whether the function is running
        self.iteration_count: int = 0
        self.last_result: typing.Optional[typing.Any] = None
        self.iteration_exception: int = 0
        self.last_exc: str = "No exception has occurred yet."
        self.last_exc_raw: typing.Optional[BaseException] = None
        self.stop: bool = False

        self.task: typing.Optional[asyncio.Task] = None
        if start_now:
            self.start()

    def start(self) -> typing.Any:  # typing_extensions.Self
        self.task = self.cog.bot.loop.create_task(self.loop())
        return self

    async def wait_until_iteration(self) -> None:
        """Sleep during the raw interval."""
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        time = now.timestamp()
        time = math.ceil(time / self.interval) * self.interval
        next_iteration = datetime.datetime.fromtimestamp(time, tz=datetime.timezone.utc) - now
        seconds_to_sleep = (next_iteration).total_seconds()
        if self.interval > 60 and hasattr(self.cog, "logger"):
            self.cog.logger.verbose(
                f"Sleeping for {seconds_to_sleep} seconds until {self.name} loop next iteration"
                f" ({self.iteration_count + 1})..."
            )
        await asyncio.sleep(seconds_to_sleep)

    async def loop(self) -> None:
        await self.cog.bot.wait_until_red_ready()
        await asyncio.sleep(1)
        if hasattr(self.cog, "logger"):
            self.cog.logger.debug(f"{self.name} loop has started.")
        while True:
            await self.execute()
            if self.maybe_stop():
                return
            if not self.wait_raw:
                # both iteration_finish and iteration_error set next_iteration as not None
                assert self.next_iteration is not None
                if float(self.interval) % float(3600) == 0:
                    self.next_iteration = self.next_iteration.replace(
                        minute=0, second=0, microsecond=0
                    )  # ensure further iterations are on the hour
                elif float(self.interval) % float(60) == 0:
                    self.next_iteration = self.next_iteration.replace(
                        second=0, microsecond=0
                    )  # ensure further iterations are on the minute
                else:
                    self.next_iteration = self.next_iteration.replace(
                        microsecond=0
                    )  # ensure further iterations are on the second
                if self.interval != 0:
                    await self.wait_until_iteration()
            else:
                await self.sleep_until_next()

    async def execute(self) -> None:
        try:
            start = time.monotonic()
            self.iteration_start()
            self.last_result: typing.Any = await self.function(**self.function_kwargs)
            self.iteration_finish()
            end = time.monotonic()
            total = round(end - start, 1)
            if hasattr(self.cog, "logger"):
                if self.iteration_count == 1:
                    self.cog.logger.verbose(
                        f"{self.name} initial iteration finished in {total}s"
                        f" ({self.iteration_count})."
                    )
                elif self.interval > 60:
                    self.cog.logger.verbose(
                        f"{self.name} iteration finished in {total}s ({self.iteration_count})."
                    )
        except Exception as e:
            if hasattr(self.cog, "logger"):
                if self.iteration_count == 1:
                    self.cog.logger.exception(
                        (
                            f"Something went wrong in the {self.name} loop"
                            f" ({self.iteration_count})."
                        ),
                        exc_info=e,
                    )
                else:
                    self.cog.logger.exception(
                        (
                            f"Something went wrong in the {self.name} loop iteration"
                            f" ({self.iteration_count})."
                        ),
                        exc_info=e,
                    )
            self.iteration_error(e)

    def maybe_stop(self) -> bool:
        if self.stop:
            return True
        if self.stop_manually:
            self.stop_all()
            return True
        if self.limit_count is not None and self.iteration_count >= self.limit_count:
            self.stop_all()
            return True
        if self.limit_date is not None and datetime.datetime.timestamp(
            datetime.datetime.now()
        ) >= datetime.datetime.timestamp(self.limit_date):
            self.stop_all()
            return True
        if self.limit_exception is not None and self.iteration_exception >= self.limit_exception:
            self.stop_all()
            return True
        return False

    def stop_all(self) -> typing.Any:  # typing_extensions.Self
        self.stop = True
        self.next_iteration = None
        self.task.cancel()
        # if self.loops.get(self.name) == self:
        #     del self.cog.loops[self.cog.loops.index(self)]
        if hasattr(self.cog, "logger"):
            self.cog.logger.debug(
                f"{self.name} loop has been stopped after {self.iteration_count} iteration(s)."
            )
        return self

    def __repr__(self) -> str:
        return (
            f"<friendly_name={self.name!r} iteration_count={self.iteration_count} "
            f"currently_running={self.currently_running} last_iteration={self.last_iteration!r} "
            f"next_iteration={self.next_iteration!r} integrity={self.integrity}>"
        )

    @property
    def integrity(self) -> bool:
        """
        If the loop is running on time (whether or not next expected iteration is in the future)
        """
        if self.next_iteration is None:  # not started yet
            return False
        return self.next_iteration > datetime.datetime.now(tz=datetime.timezone.utc)

    @property
    def until_next(self) -> float:
        """
        Positive float with the seconds until the next iteration, based off the last
        iteration and the interval.
        If the expected time of the next iteration is in the past, this will return `0.0`
        """
        if self.next_iteration is None:  # not started yet
            return 0.0

        raw_until_next = (
            self.next_iteration - datetime.datetime.now(tz=datetime.timezone.utc)
        ).total_seconds()
        if raw_until_next > self.expected_interval.total_seconds():  # should never happen
            return self.expected_interval.total_seconds()
        elif raw_until_next > 0.0:
            return raw_until_next
        else:
            return 0.0

    async def sleep_until_next(self) -> None:
        """Sleep until the next iteration. Basically an "all-in-one" version of `until_next`."""
        await asyncio.sleep(self.until_next)

    def iteration_start(self) -> None:
        """Register an iteration as starting."""
        self.iteration_count += 1
        self.currently_running = True
        self.last_iteration = datetime.datetime.now(tz=datetime.timezone.utc)
        self.next_iteration = datetime.datetime.now(tz=datetime.timezone.utc) + self.expected_interval
        # this isn't accurate, it will be "corrected" when finishing is called

    def iteration_finish(self) -> None:
        """Register an iteration as finished successfully."""
        self.currently_running = False
        # now this is accurate. imo its better to have something than nothing

    def iteration_error(self, error: BaseException) -> None:
        """Register an iteration's error."""
        self.currently_running = False
        self.last_exc_raw = error
        self.last_exc = "".join(
            traceback.format_exception(type(error), error, error.__traceback__)
        )

    def get_debug_embed(self) -> discord.Embed:
        """Get an embed with many infomations on this loop."""
        now: datetime.datetime = datetime.datetime.now(tz=datetime.timezone.utc)

        raw_table = Table("Key", "Value")
        raw_table.add_row("expected_interval", str(self.expected_interval))
        raw_table.add_row("iteration_count", str(self.iteration_count))
        raw_table.add_row("currently_running", str(self.currently_running))
        raw_table.add_row("last_iteration", str(self.last_iteration))
        raw_table.add_row("next_iteration", str(self.next_iteration))
        raw_table.add_row("wait_raw", str(self.wait_raw))
        raw_table_str = no_colour_rich_markup(raw_table, lang="py")

        if self.next_iteration is not None and self.last_iteration is not None:
            processed_table = Table("Key", "Value")
            processed_table.add_row(
                "Seconds until next", str((self.next_iteration - now).total_seconds())
            )
            processed_table.add_row(
                "Seconds since last", str((now - self.last_iteration).total_seconds())
            )
            processed_table.add_row(
                "Raw interval",
                str(
                    (self.next_iteration - now).total_seconds()
                    + (now - self.last_iteration).total_seconds()
                ),
            )
            processed_table_str = no_colour_rich_markup(processed_table, lang="py")
        elif self.iteration_count:
            processed_table_str = "Loop isn't running."
        else:
            processed_table_str = "Loop hasn't started yet."

        datetime_table = Table("Key", "Value")
        datetime_table.add_row("Start DateTime", str(self.start_datetime))
        datetime_table.add_row("Now DateTime", str(now))
        datetime_table.add_row(
            "Runtime",
            (
                str(now - self.start_datetime)
                + "\n"
                + str((now - self.start_datetime).total_seconds())
                + "s"
            ),
        )
        datetime_table_str = no_colour_rich_markup(datetime_table, lang="py")

        function_table = Table("Key", "Value")
        function_table.add_row(
            "Function", repr(getattr(self.function, "__func__", self.function))[:-23] + ">"
        )
        function_table.add_row("Function parameters", repr(inspect.signature(self.function)))
        function_table.add_row("Function kwargs", repr(self.function_kwargs))
        function_table_str = no_colour_rich_markup(function_table, lang="py")

        stopping_table = Table("Key", "Value")
        stopping_table.add_row("Limit Count", str(self.limit_count))
        stopping_table.add_row("Limit Date", str(self.limit_date))
        stopping_table.add_row("Limit Exception", str(self.limit_exception))
        stopping_table_str = no_colour_rich_markup(stopping_table, lang="py")

        emoji = "✅" if self.integrity else "❌"
        embed: discord.Embed = discord.Embed(
            title=f"{self.name} Loop: `{emoji}`",
            color=discord.Color.green() if self.integrity else discord.Color.red(),
            timestamp=now,
        )
        embed.add_field(name="Raw data:", value=raw_table_str, inline=False)
        embed.add_field(
            name="Processed data:",
            value=processed_table_str,
            inline=False,
        )
        embed.add_field(
            name="DateTime data:",
            value=datetime_table_str,
            inline=False,
        )
        embed.add_field(
            name="Function data:",
            value=function_table_str,
            inline=False,
        )
        embed.add_field(
            name="Stopping data:",
            value=stopping_table_str,
            inline=False,
        )
        exc = self.last_exc
        exc = CogsUtils.replace_var_paths(exc)
        if len(exc) > 1024:
            exc = list(pagify(exc, page_length=1024))[0] + "\n..."
        embed.add_field(name="Exception:", value=box(exc), inline=False)

        return embed



import concurrent.futures
import functools
import os
import sqlite3
from asyncio.events import AbstractEventLoop
from typing import Optional

from starbot.core.bot import StarBot
from starbot.core.data_manager import cog_data_path

# (comparisons from red config, mainly to make me feel like i didn't waste an evening)
# (these are with the stattrack cog, this driver is also used in betteruptime)
# this compares a write for config to appending to the SQL. i've not compared writes because they
# will only happend once, for migration from config. SQL includes copying DF + executor overhead
# basically this is the raw speed changes for the loop itself
# ~1.3 sec to ~0.03 sec, dataset of ~1 week on windows
# ~5-6 sec to ~0.04 sec, dataset of ~1 month on linux
# reads are insignificant as only happen on cog load

try:
    import pandas
except ImportError:
    raise RuntimeError("Pandas must be installed for this driver to work.")


class PandasSQLiteDriver:
    """An asynchronous SQLite driver for Pandas dataframes."""

    def __init__(self, bot: StarBot, cog_name: str, filename: str, table: str = "main_df") -> None:
        """Get a driver object for interacting with a table in the given cog's datapath.

        Parameters
        ----------
        bot : StarBot
            Bot object
        cog_name : str
            Full cog name, LikeThis
        filename : str
            The full file name to use for the database, for example `timeseries.db`
        table : str, optional
            The SQLite table to use, by default "main_df"
        """
        self.bot = bot
        self.table = table

        self.sql_executor = concurrent.futures.ThreadPoolExecutor(1, f"{cog_name.lower()}_sql")
        self.sql_path = str(cog_data_path(raw_name=cog_name) / filename)

    def _write(self, df: pandas.DataFrame, table: Optional[str] = None) -> None:
        connection = sqlite3.connect(self.sql_path)
        try:
            df.to_sql(table or self.table, con=connection, if_exists="replace")  # type:ignore
            connection.commit()
        finally:
            connection.close()

    def _append(self, df: pandas.DataFrame, table: Optional[str] = None) -> None:
        connection = sqlite3.connect(self.sql_path)
        try:
            df.to_sql(table or self.table, con=connection, if_exists="append")  # type:ignore
            connection.commit()
        finally:
            connection.close()

    def _read(self, table: Optional[str] = None) -> pandas.DataFrame:
        connection = sqlite3.connect(self.sql_path)
        try:
            df = pandas.read_sql(
                f"SELECT * FROM {table or self.table}",
                connection,
                index_col="index",
                parse_dates=["index"],
            )
            return df
        finally:
            connection.close()

    async def write(self, df: pandas.DataFrame, table: Optional[str] = None) -> None:
        """Write a dataframe to the database. Replaces and old data."""
        assert isinstance(self.bot.loop, AbstractEventLoop)
        func = functools.partial(self._write, df.copy(True), table)
        await self.bot.loop.run_in_executor(self.sql_executor, func)

    async def append(self, df: pandas.DataFrame, table: Optional[str] = None) -> None:
        """Append a dataframe to the database."""
        assert isinstance(self.bot.loop, AbstractEventLoop)
        func = functools.partial(self._append, df.copy(True), table)
        await self.bot.loop.run_in_executor(self.sql_executor, func)

    async def read(self, table: Optional[str] = None) -> pandas.DataFrame:
        """Read the database, returning as a pandas dataframe."""
        assert isinstance(self.bot.loop, AbstractEventLoop)
        func = functools.partial(self._read, table)
        return await self.bot.loop.run_in_executor(self.sql_executor, func)

    def storage_usage(self) -> int:
        """Return the size of the database file in bytes."""
        return os.path.getsize(self.sql_path)



# Menu Utilities for Star-Cogs

from starbot.core import commands  # isort:skip
from starbot.core.bot import StarBot  # isort:skip
import discord  # isort:skip
import typing  # isort:skip

import asyncio
import re

from colorama import Fore
from starbot.core.utils.chat_formatting import box, pagify, text_to_file
from starbot.core.utils.menus import start_adding_reactions
from starbot.core.utils.predicates import MessagePredicate, ReactionPredicate
from starbot.vendored.discord.ext import menus

from .cogsutils import CogsUtils

__all__ = ["Menu", "Reactions"]


def _(untranslated: str) -> str:
    return untranslated


def cleanup_ansi(text: str) -> str:
    for attr in dir(Fore):
        if attr.startswith("_") or not isinstance(getattr(Fore, attr), str):
            continue
        text = text.replace(getattr(Fore, attr), "")
    return text


class Menu(discord.ui.View):
    """Create Menus easily."""

    def __init__(
        self,
        pages: typing.List[
            typing.Union[typing.Dict[str, typing.Union[str, typing.Any]], discord.Embed, str]
        ],
        timeout: typing.Optional[int] = 180,
        delete_after_timeout: typing.Optional[bool] = False,
        page_start: typing.Optional[int] = 0,
        members: typing.Optional[typing.Iterable[typing.Union[discord.Member, int]]] = None,
        ephemeral: typing.Optional[bool] = False,
        prefix: typing.Optional[str] = None,
        lang: typing.Optional[str] = None,
    ) -> None:
        if members is None:
            members = []
        super().__init__(timeout=timeout)
        self.ctx: commands.Context = None
        self.pages: typing.List[
            typing.Union[str, discord.Embed, typing.Dict[str, typing.Any]]
        ] = pages
        self.delete_after_timeout: bool = delete_after_timeout
        controls: typing.Dict[str, str] = {
            "⏮️": "left_page",
            "◀️": "prev_page",
            "✖️": "close_page",
            "▶️": "next_page",
            "⏭️": "right_page",
            "🔻": "send_all",
            "📩": "send_interactive",
            "💾": "send_as_file",
            "choose_page": "choose_page",
        }
        self.controls: typing.Dict[str, str] = controls.copy()
        self.extra_items: typing.List[discord.ui.Item] = []
        self.members: typing.Optional[typing.List[int]] = (
            members if members is None else [getattr(member, "id", member) for member in members]
        )
        self.ephemeral: bool = ephemeral
        if not self.pages:
            self.pages: typing.List[str] = ["Nothing to show."]
        self.prefix: typing.Optional[str] = prefix
        if self.prefix is not None and not self.prefix.startswith("```"):
            self.prefix = box(self.prefix, lang="py")
        if isinstance(self.pages, str):
            self.pages: typing.List[str] = list(
                pagify(
                    self.pages,
                    shorten_by=len(f"```{lang or ''}\n\n```")
                    + (len(f"{self.prefix}\n") if self.prefix is not None else 0),
                )
            )
        self.lang: typing.Optional[str] = lang
        if (self.prefix is not None or lang is not None) and all(
            isinstance(page, str) for page in self.pages
        ):
            self.pages: typing.List[str] = [
                (self.prefix or "")
                + "\n"
                + (
                    box(page[: 2000 - len(f"```{self.lang}\n\n```")], self.lang)
                    if self.lang is not None
                    else page
                )
                for page in self.pages
            ]
        if not isinstance(self.pages[0], (typing.Dict, discord.Embed, str)):
            raise RuntimeError("Pages must be of type typing.Dict, discord.Embed or str.")
        if len(self.pages) > 1:
            self.disabled_controls: typing.List[str] = []
        else:
            self.disabled_controls: typing.List[str] = ["send_as_file", "choose_page"]

        self._source: self._SimplePageSource = self._SimplePageSource(items=self.pages)
        if not self._source.is_paginating():
            for emoji, name in controls.items():
                if name in ("left_page", "prev_page", "next_page", "right_page"):
                    del self.controls[emoji]
                    self.disabled_controls.append(name)
        if not self._source.is_paginating() or len(self.pages) > 3:
            for emoji, name in controls.items():
                if name in ("send_all"):
                    del self.controls[emoji]
                    self.disabled_controls.append(name)
        if not self._source.is_paginating() or len(self.pages) <= 3:
            for emoji, name in controls.items():
                if name in ("send_interactive"):
                    del self.controls[emoji]
                    self.disabled_controls.append(name)
        if not all(isinstance(page, str) for page in self.pages):
            for emoji, name in controls.items():
                if name in ("send_as_file"):
                    del self.controls[emoji]
                    self.disabled_controls.append(name)

        self._message: discord.Message = None
        self._current_page: int = page_start
        self._is_done = asyncio.Event()

    async def start(self, ctx: commands.Context, wait: bool = True) -> None:
        """
        Used to start the menu displaying the first page requested.
        Parameters
        ----------
            ctx: `commands.Context`
                The context to start the menu in.
        """
        self.ctx: commands.Context = ctx
        current, kwargs = await self.get_page(self._current_page)
        if not self.ctx.channel.permissions_for(self.ctx.author).send_messages:
            for emoji, name in self.controls.copy().items():
                if name in ("send_all", "send_interactive"):
                    del self.controls[emoji]
                    self.disabled_controls.append(name)
        for button in self.children:
            if button.custom_id in self.disabled_controls:
                self.remove_item(button)
        for button in self.children:
            self.remove_item(button)
            self.add_item(button)
        for item in self.extra_items:
            self.add_item(item)
        if choose_button := discord.utils.get(self.children, custom_id="choose_page"):
            choose_button.label = f"Page {current + 1}/{len(self.pages)}"
        if self.ephemeral and self.ctx.interaction is not None:
            kwargs["ephemeral"] = True
        if self._message is None:
            self._message = await ctx.send(**kwargs, view=self)
        else:
            self._current_page = 0
            await self.change_page()
        for page in self.pages:
            if isinstance(page, typing.Dict) and "file" in page:
                # del page["file"]  # ValueError: I/O operation on closed file.
                page["attachments"] = [page.pop("file")]
        if wait:
            await self._is_done.wait()
        return self._message

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        if interaction.user.id not in [self.ctx.author.id] + self.members + list(
            self.ctx.bot.owner_ids
        ):
            await interaction.response.send_message(
                "You are not allowed to use this interaction.", ephemeral=True
            )
            return False
        return True

    async def on_timeout(self) -> None:
        if not self.delete_after_timeout:
            for child in self.children:
                child: discord.ui.Item
                if hasattr(child, "disabled") and not (
                    isinstance(child, discord.ui.Button) and child.style == discord.ButtonStyle.url
                ):
                    child.disabled = True
            try:
                await self._message.edit(view=self)
            except discord.HTTPException:
                pass
        else:
            try:
                await self._message.delete()
            except discord.HTTPException:
                pass
        self.stop()
        self._is_done.set()

    async def get_page(
        self, page_num: int
    ) -> typing.Dict[str, typing.Union[str, discord.Embed, typing.Any]]:
        try:
            page = await self._source.get_page(page_num)
        except IndexError:
            self.current_page = 0
            page = await self._source.get_page(self.current_page)
        current = self.pages.index(page)
        value = await self._source.format_page(self, page)

        if isinstance(value, typing.Dict):
            if "content" in value:
                value["content"] = CogsUtils.replace_var_paths(value["content"])
            return current, value
        elif isinstance(value, str):
            value = CogsUtils.replace_var_paths(value)
            return current, {"content": value, "embed": None}
        elif isinstance(value, discord.Embed):
            return current, {"embed": value, "content": None}

    async def change_page(self, interaction: typing.Optional[discord.Interaction] = None) -> None:
        if interaction is not None:
            try:
                await interaction.response.defer()
            except discord.HTTPException:
                pass
        current, kwargs = await self.get_page(self._current_page)
        if choose_button := discord.utils.get(self.children, custom_id="choose_page"):
            choose_button.label = f"Page {current + 1}/{len(self.pages)}"
        self._message = await self._message.edit(**kwargs, view=self)

    @discord.ui.button(emoji="⏮️", custom_id="left_page")
    async def left_page(self, interaction: discord.Interaction, button: discord.ui.Button) -> None:
        self._current_page = 0
        await self.change_page(interaction)

    @discord.ui.button(emoji="◀️", custom_id="prev_page")
    async def prev_page(self, interaction: discord.Interaction, button: discord.ui.Button) -> None:
        self._current_page += -1
        await self.change_page(interaction)

    @discord.ui.button(emoji="✖️", style=discord.ButtonStyle.danger, custom_id="close_page")
    async def close_page(
        self, interaction: discord.Interaction, button: discord.ui.Button
    ) -> None:
        await interaction.response.defer()
        try:
            await self._message.delete()
        except discord.HTTPException:
            pass
        self.stop()
        self._is_done.set()

    @discord.ui.button(emoji="▶️", custom_id="next_page")
    async def next_page(self, interaction: discord.Interaction, button: discord.ui.Button) -> None:
        self._current_page += 1
        await self.change_page(interaction)

    @discord.ui.button(emoji="⏭️", custom_id="right_page")
    async def right_page(
        self, interaction: discord.Interaction, button: discord.ui.Button
    ) -> None:
        self._current_page = self._source.get_max_pages() - 1
        await self.change_page(interaction)

    @discord.ui.button(emoji="🔻", custom_id="send_all")
    async def send_all(self, interaction: discord.Interaction, button: discord.ui.Button) -> None:
        await interaction.response.defer()
        for i in range(len(self.pages)):
            __, kwargs = await self.get_page(i)
            if self.prefix is not None and i != 0 and kwargs["content"] is not None:
                kwargs["content"] = kwargs["content"][len(self.prefix) :]
            await interaction.channel.send(**kwargs)

    @discord.ui.button(emoji="📩", custom_id="send_interactive")
    async def send_interactive(
        self, interaction: discord.Interaction, button: discord.ui.Button
    ) -> None:
        await interaction.response.defer()
        ret = []
        for i in range(len(self.pages)):
            current, kwargs = await self.get_page(i)
            if self.prefix is not None and i != 0 and kwargs["content"] is not None:
                kwargs["content"] = kwargs["content"][len(self.prefix) :]
            msg = await self.ctx.send(**kwargs)
            ret.append(msg)
            n_remaining = len(self.pages) - current
            if n_remaining <= 0:
                break
            elif n_remaining == 1:
                plural = ""
                is_are = "is"
            else:
                plural = "s"
                is_are = "are"
            query = await self.ctx.send(
                f"There {is_are} still {n_remaining} message{plural} remaining. Type `more` to"
                " continue."
            )
            try:
                resp = await self.ctx.bot.wait_for(
                    "message",
                    check=MessagePredicate.lower_equal_to("more", self.ctx),
                    timeout=15,
                )
            except asyncio.TimeoutError:
                try:
                    await query.delete()
                except discord.HTTPException:
                    pass
                break
            else:
                try:
                    await self.ctx.channel.delete_messages((query, resp))
                except (discord.HTTPException, AttributeError):
                    try:
                        await query.delete()
                    except discord.HTTPException:
                        pass
                continue
        return ret

    @discord.ui.button(emoji="💾", custom_id="send_as_file")
    async def send_as_file(
        self, interaction: discord.Interaction, button: discord.ui.Button
    ) -> None:
        await interaction.response.defer(ephemeral=True, thinking=True)

        def cleanup_code(content):
            """Automatically removes code blocks from the code."""
            # remove ˋˋˋpy\n````
            if content.startswith("```") and content.endswith("```"):
                content = re.compile(r"^((```py(thon)?)(?=\s)|(```))").sub("", content)[:-3]
            return content.strip("` \n")

        if not self.ctx.channel.permissions_for(self.ctx.me).attach_files:
            await interaction.response.send_message(
                _("I don't have the permission to attach files in this channel."), ephemeral=True
            )
            return
        pages = self.pages.copy()
        if self.prefix is not None:
            for i, page in enumerate(pages):
                pages[i] = page[len(self.prefix) :]
        all_text = [cleanup_code(cleanup_ansi(page)) for page in pages]
        all_text = (f"{self.prefix}\n\n" if self.prefix is not None else "") + "\n".join(all_text)
        await interaction.followup.send(
            file=text_to_file(
                all_text,
                filename=f"Menu_{interaction.message.channel.id}-{interaction.message.id}.txt",
            ),
            ephemeral=self.ephemeral,
        )

    @discord.ui.button(label="Page 1/1", custom_id="choose_page")
    async def choose_page(
        self, interaction: discord.Interaction, button: discord.ui.Button
    ) -> None:
        class ChoosePageModal(discord.ui.Modal):
            def __init__(_self):
                super().__init__(title="Choose page")
                nb_pages = len(self.pages)
                _self.item = discord.ui.TextInput(
                    label=f"Page number between 1 and {nb_pages}",
                    placeholder="Page number",
                    required=True,
                    max_length=len(str(nb_pages)),
                )
                _self.add_item(_self.item)

            async def on_submit(_self, interaction: discord.Interaction):
                # Too late.
                if self._is_done.is_set():
                    await interaction.response.send_message(
                        _("Too late. The Menu is already finished."),
                        ephemeral=True,
                    )
                    return
                # Integer.
                try:
                    page = int(_self.item.value)
                except ValueError:
                    await interaction.response.send_message(
                        _("The page number must be an int."),
                        ephemeral=True,
                    )
                    return
                # Minimum-Maximum.
                _max = len(self.pages)
                if page < 1 or page > _max:
                    await interaction.response.send_message(
                        _("The page number must be between 1 and {_max}.").format(_max=_max),
                        ephemeral=True,
                    )
                    return
                # Change the page.
                self._current_page = page - 1
                await self.change_page(interaction)

        modal = ChoosePageModal()
        await interaction.response.send_modal(modal)

    class _SimplePageSource(menus.ListPageSource):
        def __init__(
            self,
            items: typing.List[
                typing.Union[
                    typing.Dict[str, typing.Union[str, discord.Embed]], discord.Embed, str
                ]
            ],
        ) -> None:
            super().__init__(items, per_page=1)

        async def format_page(
            self,
            view,
            page: typing.Union[
                typing.Dict[str, typing.Union[str, discord.Embed]], discord.Embed, str
            ],
        ) -> typing.Union[str, discord.Embed]:
            return page


class Reactions:
    """Create Reactions easily."""

    def __init__(
        self,
        bot: StarBot,
        message: discord.Message,
        remove_reaction: typing.Optional[bool] = True,
        timeout: typing.Optional[int] = 180,
        reactions: typing.Optional[typing.List] = None,
        members: typing.Optional[typing.Iterable[typing.Union[discord.Member, int]]] = None,
        check: typing.Optional[typing.Callable] = None,
        function: typing.Optional[typing.Callable] = None,
        function_args: typing.Optional[typing.Dict] = None,
        infinity: typing.Optional[bool] = False,
    ) -> None:
        if reactions is None:
            reactions = ["✅", "❌"]
        if function_args is None:
            function_args = {}
        self.reactions_dict_instance: typing.Dict[str, typing.Any] = {
            "message": message,
            "timeout": timeout,
            "reactions": reactions,
            "members": members,
            "check": check,
            "function": function,
            "function_args": function_args,
            "infinity": infinity,
        }
        self.bot: StarBot = bot
        self.message: discord.Message = message
        self.remove_reaction: bool = remove_reaction
        self.timeout: int = timeout
        self.infinity: bool = infinity
        self.reaction_result: typing.Union[str, discord.PartialEmoji] = None
        self.user_result: discord.User = None
        self.function_result: typing.Optional[typing.Any] = None
        self.members: typing.Optional[typing.List[int]] = (
            members if members is None else [getattr(member, "id", member) for member in members]
        )
        self.check: typing.Optional[typing.Callable] = check
        self.function: typing.Optional[typing.Callable] = function
        self.function_args: typing.Optional[typing.Dict[str, typing.Any]] = function_args
        self.reactions: typing.List[str] = reactions
        self.r: bool = False
        self.done: asyncio.Event = asyncio.Event()
        asyncio.create_task(self.wait())

    def to_dict_cogsutils(
        self, for_Config: typing.Optional[bool] = False
    ) -> typing.Dict[str, typing.Any]:
        reactions_dict_instance = self.reactions_dict_instance
        if for_Config:
            reactions_dict_instance["bot"] = None
            reactions_dict_instance["message"] = None
            reactions_dict_instance["members"] = None
            reactions_dict_instance["check"] = None
            reactions_dict_instance["function"] = None
        return reactions_dict_instance

    @classmethod
    def from_dict_cogsutils(
        cls, reactions_dict_instance: typing.Dict
    ) -> typing.Any:  # typing_extensions.Self
        return cls(**reactions_dict_instance)

    async def wait(self) -> None:
        if not self.r:
            await start_adding_reactions(self.message, self.reactions)
            self.r = True
        predicates = ReactionPredicate.same_context(message=self.message)
        running = True
        try:
            while running:
                tasks = [asyncio.create_task(self.bot.wait_for("reaction_add", check=predicates))]
                done, pending = await asyncio.wait(
                    tasks, timeout=self.timeout, return_when=asyncio.FIRST_COMPLETED
                )
                for task in pending:
                    task.cancel()
                if len(done) == 0:
                    raise TimeoutError()
                reaction, user = done.pop().result()
                running = await self.reaction_check(reaction, user)
        except TimeoutError:
            await self.on_timeout()

    async def reaction_check(self, reaction: discord.Reaction, user: discord.User) -> bool:
        async def remove_reaction(
            remove_reaction,
            message: discord.Message,
            reaction: discord.Reaction,
            user: discord.User,
        ) -> None:
            if remove_reaction:
                try:
                    await message.remove_reaction(emoji=reaction, member=user)
                except discord.HTTPException:
                    pass

        if str(reaction.emoji) not in self.reactions:
            await remove_reaction(self.remove_reaction, self.message, reaction, user)
            return False
        if self.members is not None and user.id not in self.members:
            await remove_reaction(self.remove_reaction, self.message, reaction, user)
            return False
        if self.check is not None and not self.check(reaction, user):
            await remove_reaction(self.remove_reaction, self.message, reaction, user)
            return False
        await remove_reaction(self.remove_reaction, self.message, reaction, user)
        self.reaction_result = reaction
        self.user_result = user
        if self.function is not None:
            self.function_result = await self.function(self, reaction, user, **self.function_args)
        self.done.set()
        return self.infinity

    async def on_timeout(self) -> None:
        self.done.set()

    async def wait_result(
        self,
    ) -> typing.Tuple[
        typing.Union[discord.PartialEmoji, str], discord.User, typing.Optional[typing.Any]
    ]:
        self.done = asyncio.Event()
        await self.done.wait()
        reaction, user, function_result = self.get_result()
        if reaction is None:
            raise TimeoutError()
        self.reaction_result, self.user_result, self.function_result = None, None, None
        return reaction, user, function_result

    def get_result(
        self,
    ) -> typing.Tuple[
        typing.Union[discord.PartialEmoji, str], discord.User, typing.Optional[typing.Any]
    ]:
        return self.reaction_result, self.user_result, self.function_result



from typing import TYPE_CHECKING, Optional

import discord
from starbot.core.bot import StarBot

from .chat import humanize_bytes, inline_hum_list, no_colour_rich_markup
from .meta import format_info, get_vex_logger, out_of_date_check
from .__version__ import __version__
from . import cog
from .cog import Cog
from .cogsutils import CogsUtils
from .context import Context
from .loop import Loop
from .menus import Menu, Reactions
from .sentry import SentryHelper
from .settings import Settings
from .shared_cog import SharedCog
from .views import (
    Buttons,
    ChannelSelect,
    ConfirmationAskView,
    Dropdown,
    MentionableSelect,
    Modal,
    RoleSelect,
    Select,
    UserSelect,
)  # NOQA

cog.SharedCog = SharedCog

__author__ = "Scarlet"
__version__ = __version__

__all__ = [
    "CogsUtils",
    "Loop",
    "SharedCog",
    "Cog",
    "Menu",
    "Context",
    "Settings",
    "SentryHelper",
    "ConfirmationAskView",
    "Buttons",
    "Dropdown",
    "Select",
    "ChannelSelect",
    "MentionableSelect",
    "RoleSelect",
    "UserSelect",
    "Modal",
    "Reactions",
    "humanize_bytes",
    "inline_hum_list",
    "no_colour_rich_markup",
    "format_info",
    "get_vex_logger",
    "out_of_date_check",
]


from starbot.core import commands  # isort:skip
import discord  # isort:skip
import typing  # isort:skip
from typing import Any, List, Optional, Union, Literal, Dict, Iterable, Tuple

import asyncio
import inspect
import string
from dataclasses import dataclass
from functools import partial
from random import choice

from discord import ButtonStyle, Embed, Interaction, ui, SelectOption, ComponentType
from discord.http import Route
from starbot.core.bot import StarBot

__all__ = [
    # Utility classes and functions
    "PredItem",
    "wait_for_press",
    "wait_for_yes_no",
    "URLButton",
    "send_message",
    # View components
    "ConfirmationAskView",
    "Buttons",
    "Dropdown",
    "Select",
    "ChannelSelect",
    "MentionableSelect",
    "RoleSelect",
    "UserSelect",
    "Modal",
]


def _(untranslated: str) -> str:
    return untranslated


def generate_key(length: typing.Optional[int] = 10) -> str:
    """Generate a random string of lowercase letters and digits."""
    strings = []
    strings += string.ascii_lowercase
    strings += string.digits
    return "".join(choice(strings) for _ in range(length))


# ---------- Button-Based Interaction Utilities ----------

@dataclass
class PredItem:
    ref: Any
    style: ButtonStyle
    label: str
    row: Optional[int] = None


class _PredView(ui.View):
    def __init__(self, timeout: Optional[float], author_id: int):
        super().__init__(timeout=timeout)
        self.ref: Any = None
        self.author_id = author_id
        self.pressed = asyncio.Event()

    async def interaction_check(self, interaction: Interaction) -> bool:
        if interaction.user.id == self.author_id:
            return True
        await interaction.response.send_message(
            "You don't have permission to do this.", ephemeral=True
        )
        return False


class _PredButton(ui.Button):
    def __init__(self, ref: Any, style: ButtonStyle, label: str, row: Optional[int] = None):
        super().__init__(style=style, label=label, row=row)
        self.ref = ref

    async def callback(self, interaction: Interaction):
        assert isinstance(self.view, _PredView)
        self.view.stop()
        self.view.ref = self.ref
        self.view.pressed.set()


async def wait_for_press(
    ctx: commands.Context,
    items: List[PredItem],
    content: Optional[str] = None,
    embed: Optional[Embed] = None,
    *,
    timeout: float = 180.0,
) -> Any:
    """Wait for a button press from a list of options.
    
    Args:
        ctx: Command context
        items: List of PredItem objects representing buttons
        content: Message content
        embed: Message embed
        timeout: Timeout in seconds
        
    Returns:
        The reference value of the button that was pressed
    """
    if not items:
        raise ValueError("The `items` argument cannot be empty.")

    view = _PredView(timeout, ctx.author.id)
    for i in items:
        button = _PredButton(i.ref, i.style, i.label, i.row)
        view.add_item(button)
    message = await ctx.send(content=content, embed=embed, view=view)

    await asyncio.wait_for(view.pressed.wait(), timeout=timeout)

    # Disable buttons after interaction
    emptyview = ui.View()
    for i in items:
        emptyview.add_item(
            ui.Button(
                style=i.style if i.ref == view.ref else ButtonStyle.gray,
                label=i.label,
                row=i.row,
                disabled=True,
            )
        )
    await message.edit(view=emptyview)
    emptyview.stop()

    return view.ref


async def wait_for_yes_no(
    ctx: commands.Context,
    content: Optional[str] = None,
    embed: Optional[Embed] = None,
    *,
    timeout: float = 180.0,
) -> bool:
    """Wait for a Yes/No button press.
    
    Args:
        ctx: Command context
        content: Message content
        embed: Message embed
        timeout: Timeout in seconds
        
    Returns:
        True if "Yes" was pressed, False if "No" was pressed
    """
    view = _PredView(timeout, ctx.author.id)
    view.add_item(_PredButton(True, ButtonStyle.blurple, "Yes"))
    view.add_item(_PredButton(False, ButtonStyle.blurple, "No"))

    message = await ctx.send(content=content, embed=embed, view=view)

    await asyncio.wait_for(view.pressed.wait(), timeout=timeout)

    # Disable buttons after interaction
    emptyview = ui.View()
    emptyview.add_item(
        ui.Button(
            style=ButtonStyle.grey if view.ref is False else ButtonStyle.blurple,
            label="Yes",
            disabled=True,
        )
    )
    emptyview.add_item(
        ui.Button(
            style=ButtonStyle.grey if view.ref is True else ButtonStyle.blurple,
            label="No",
            disabled=True,
        )
    )
    await message.edit(view=emptyview)
    emptyview.stop()

    return view.ref


# ---------- URL Button Implementation ----------

class URLButton:
    """A simple URL button that can be used without a View."""
    
    def __init__(self, label: str, url: str) -> None:
        if not isinstance(label, str):
            raise TypeError("Label must be a string")
        if not isinstance(url, str):
            raise TypeError("URL must be a string")
        self.label = label
        self.url = url

    def to_dict(self) -> dict:
        return {
            "label": self.label,
            "style": 5,
            "type": 2,
            "url": self.url,
        }


async def send_message(
    bot: StarBot,
    channel_id: int,
    *,
    content: Optional[str] = None,
    embed: Optional[discord.Embed] = None,
    file: Optional[discord.File] = None,
    url_button: Optional[URLButton] = None,
):
    """Send a message with optional file and URL button.
    
    This uses the HTTP API directly for more flexibility than channel.send().
    """
    payload = {}

    if content:
        payload["content"] = content
    if embed:
        payload["embed"] = embed.to_dict()
    if url_button:
        payload["components"] = [{"type": 1, "components": [url_button.to_dict()]}]

    if file:
        form = [
            {
                "name": "file",
                "value": file.fp,
                "filename": file.filename,
                "content_type": "application/octet-stream",
            },
            {"name": "payload_json", "value": discord.utils.to_json(payload)},
        ]
        r = Route("POST", "/channels/{channel_id}/messages", channel_id=channel_id)
        await bot._connection.http.request(r, form=form, files=[file])
    else:
        r = Route("POST", "/channels/{channel_id}/messages", channel_id=channel_id)
        await bot._connection.http.request(r, json=payload)


# ---------- Comprehensive UI Components ----------

class ConfirmationAskView(discord.ui.View):
    """Request a confirmation by the user!"""

    def __init__(
        self,
        ctx: commands.Context,
        timeout: typing.Optional[int] = 60,
        timeout_message: typing.Optional[str] = _("Timed out, please try again."),
        delete_message: typing.Optional[bool] = True,
        delete_after_timeout: typing.Optional[bool] = True,
        members: typing.Optional[typing.Iterable[typing.Union[discord.Member, int]]] = None,
    ):
        super().__init__(timeout=timeout)
        self.ctx: commands.Context = ctx

        self.delete_message: bool = delete_message
        self.timeout_message: str = timeout_message
        self.delete_after_timeout: bool = delete_after_timeout
        self.members: typing.Optional[typing.List[int]] = (
            [] if members is None else [getattr(member, "id", member) for member in members]
        )

        self._message: discord.Message = None
        self._result: typing.Optional[bool] = None

    async def start(self, *args, **kwargs) -> bool:
        """
        Request a confirmation by the user.
        """
        self._message = await self.ctx.send(*args, **kwargs, view=self)
        task_result = await self.wait()
        return None if task_result is True or self._result is None else self._result

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        if interaction.user.id not in [self.ctx.author.id] + self.members + list(
            self.ctx.bot.owner_ids
        ):
            await interaction.response.send_message(
                "You are not allowed to use this interaction.", ephemeral=True
            )
            return False
        return True

    async def on_timeout(self) -> None:
        await self.ctx.send(self.timeout_message)
        if not self.delete_after_timeout:
            for child in self.children:
                child: discord.ui.Item
                if isinstance(child, discord.ui.Button) and child.style != discord.ButtonStyle.url:
                    child.disabled = True
            try:
                await self._message.edit(view=self)
            except discord.HTTPException:
                pass
        else:
            try:
                await self._message.delete()
            except discord.HTTPException:
                pass
        self.stop()

    @discord.ui.button(
        label=_("No"), emoji="✖️", style=discord.ButtonStyle.danger, custom_id="false_button"
    )
    async def false_button(self, interaction: discord.Interaction, button: discord.ui.Button):
        await interaction.response.defer()
        if self.delete_message:
            try:
                await self._message.delete()
            except discord.HTTPException:
                pass
        self._result = False
        self.stop()

    @discord.ui.button(
        label=_("Yes"), emoji="✅", style=discord.ButtonStyle.success, custom_id="true_button"
    )
    async def true_button(self, interaction: discord.Interaction, button: discord.ui.Button):
        await interaction.response.defer()
        if self.delete_message:
            try:
                await self._message.delete()
            except discord.HTTPException:
                pass
        self._result = True
        self.stop()


class Buttons(discord.ui.View):
    """Create Buttons easily."""

    def __init__(
        self,
        timeout: typing.Optional[int] = 180,
        buttons: typing.Optional[typing.List] = None,
        members: typing.Optional[typing.Iterable[typing.Union[discord.Member, int]]] = None,
        check: typing.Optional[typing.Callable] = None,
        function: typing.Optional[typing.Callable] = None,
        function_kwargs: typing.Optional[typing.Dict[str, typing.Any]] = None,
        infinity: typing.Optional[bool] = False,
    ) -> None:
        """style: ButtonStyle, label: Optional[str], disabled: bool, custom_id: Optional[str], url: Optional[str], emoji: Optional[Union[str, Emoji, PartialEmoji]], row: Optional[int]"""
        if buttons is None:
            buttons = [{}]
        if function_kwargs is None:
            function_kwargs = {}
        for button_dict in buttons:
            if "url" in button_dict and button_dict["url"] is not None:
                button_dict["style"] = 5
                continue
            if "custom_id" not in button_dict:
                button_dict["custom_id"] = "CogsUtils" + "_" + generate_key(length=10)
        self.buttons_dict_instance: typing.Dict[str, typing.Any] = {
            "timeout": timeout,
            "buttons": [b.copy() for b in buttons],
            "members": members,
            "check": check,
            "function": function,
            "function_kwargs": function_kwargs,
            "infinity": infinity,
        }
        super().__init__(timeout=timeout)
        self.infinity: bool = infinity
        self.interaction_result: typing.Optional[discord.Interaction] = None
        self.function_result: typing.Optional[typing.Any] = None
        self.members: typing.Optional[typing.List[int]] = (
            members if members is None else [getattr(member, "id", member) for member in members]
        )
        self.check: typing.Optional[typing.Callable] = check
        self.function: typing.Optional[typing.Callable] = function
        self.function_kwargs: typing.Optional[typing.Dict[str, typing.Any]] = function_kwargs
        self.clear_items()
        self.buttons: typing.List[discord.ui.Button] = []
        self.buttons_dict: typing.List[typing.Dict[str, typing.Any]] = []
        for button_dict in buttons:
            if "style" not in button_dict:
                button_dict["style"] = discord.ButtonStyle(2)
            elif isinstance(button_dict["style"], int):
                button_dict["style"] = discord.ButtonStyle(button_dict["style"])
            if "disabled" not in button_dict:
                button_dict["disabled"] = False
            if "label" not in button_dict and "emoji" not in button_dict:
                button_dict["label"] = "Test"
            button = discord.ui.Button(**button_dict)
            self.add_item(button)
            self.buttons.append(button)
            self.buttons_dict.append(button_dict)
        self.done: asyncio.Event = asyncio.Event()

    def to_dict_cogsutils(
        self, for_Config: typing.Optional[bool] = False
    ) -> typing.Dict[str, typing.Any]:
        buttons_dict_instance = self.buttons_dict_instance
        if for_Config:
            buttons_dict_instance["check"] = None
            buttons_dict_instance["function"] = None
        return buttons_dict_instance

    @classmethod
    def from_dict_cogsutils(
        cls, buttons_dict_instance: typing.Dict
    ) -> typing.Any:  # typing_extensions.Self
        if "function_args" in buttons_dict_instance:
            buttons_dict_instance["function_kwargs"] = buttons_dict_instance["function_args"]
            del buttons_dict_instance["function_args"]
        return cls(**buttons_dict_instance)

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        if self.members is not None and interaction.user.id not in self.members:
            await interaction.response.send_message(
                "You are not allowed to use this interaction.", ephemeral=True
            )
            return False
        if self.check is not None and not self.check(interaction):
            await interaction.response.send_message(
                "You are not allowed to use this interaction.", ephemeral=True
            )
            return False
        self.interaction_result = interaction
        if self.function is not None:
            self.function_result = await self.function(self, interaction, **self.function_kwargs)
        self.done.set()
        if not self.infinity:
            self.stop()
        return True

    async def on_timeout(self) -> None:
        self.done.set()
        self.stop()

    async def on_error(self, interaction: discord.Interaction, error: Exception, item) -> None:
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        await interaction.followup.send("Sorry. An error has occurred.", ephemeral=True)
        discord.ui.view._log.error(
            "Ignoring exception in view %r for item %r.", self, item, exc_info=error
        )

    async def wait_result(self) -> typing.Tuple[discord.Interaction, typing.Optional[typing.Any]]:
        self.done = asyncio.Event()
        await self.done.wait()
        interaction, function_result = self.get_result()
        if interaction is None:
            raise TimeoutError()
        self.interaction_result, self.function_result = None, None
        return interaction, function_result

    def get_result(self) -> typing.Tuple[discord.Interaction, typing.Optional[typing.Any]]:
        return self.interaction_result, self.function_result


class Dropdown(discord.ui.View):
    """Create Dropdown easily."""

    def __init__(
        self,
        timeout: typing.Optional[int] = 180,
        placeholder: typing.Optional[str] = "Choose an option.",
        min_values: typing.Optional[int] = 1,
        max_values: typing.Optional[int] = 1,
        *,
        _type: typing.Optional[discord.ComponentType] = discord.ComponentType.select,
        options: typing.Optional[
            typing.Union[typing.List, discord.ComponentType, discord.ui.Select]
        ] = None,
        disabled: typing.Optional[bool] = False,
        members: typing.Optional[typing.Iterable[typing.Union[discord.Member, int]]] = None,
        check: typing.Optional[typing.Callable] = None,
        function: typing.Optional[typing.Callable] = None,
        function_kwargs: typing.Optional[typing.Dict[str, typing.Any]] = None,
        infinity: typing.Optional[bool] = False,
        custom_id: typing.Optional[str] = f"CogsUtils_{generate_key(length=10)}",
    ) -> None:
        """label: str, value: str, description: Optional[str], emoji: Optional[Union[str, Emoji, PartialEmoji]], default: bool"""
        if options is None:
            options = [{}]
        if function_kwargs is None:
            function_kwargs = {}
        self.dropdown_dict_instance: typing.Dict[str, typing.Any] = {
            "timeout": timeout,
            "placeholder": placeholder,
            "min_values": min_values,
            "max_values": max_values,
            "_type": _type,
            "options": ([(o.copy() if hasattr(o, "copy") else o) for o in options])
            if isinstance(options, typing.List)
            else options,
            "members": members,
            "check": check,
            "function": function,
            "function_kwargs": function_kwargs,
            "infinity": infinity,
            "custom_id": custom_id,
        }
        super().__init__(timeout=timeout)
        self._type: discord.ComponentType = _type
        self.infinity: bool = infinity
        self.interaction_result: typing.Optional[discord.Interaction] = None
        self.options_result: typing.Optional[typing.List[str]] = None
        self.function_result: typing.Optional[typing.Any] = None
        self.disabled: bool = disabled
        self.members: typing.Optional[typing.List[int]] = (
            members if members is None else [getattr(member, "id", member) for member in members]
        )
        self.check: typing.Optional[typing.Callable] = check
        self.function: typing.Optional[typing.Callable] = function
        self.function_kwargs: typing.Optional[typing.Dict[str, typing.Any]] = function_kwargs
        self.clear_items()
        self.options: typing.List[discord.SelectOption] = []
        self.options_dict: typing.List[typing.Dict[str, typing.Any]] = []
        if _type is discord.ComponentType.select or _type is Select:
            for option_dict in options:
                if "label" not in option_dict and "emoji" not in option_dict:
                    option_dict["label"] = "Test"
                option = discord.SelectOption(**option_dict)
                self.options.append(option)
                self.options_dict.append(option_dict)
            self.dropdown: discord.ui.Select = Select(
                placeholder=placeholder,
                min_values=min_values,
                max_values=max_values,
                options=self.options,
                disabled=disabled,
                custom_id=custom_id,
            )
        elif _type is discord.ComponentType.channel_select or _type is ChannelSelect:
            if options in [[{}], []]:
                options = None
            self.dropdown: discord.ui.Select = ChannelSelect(
                placeholder=placeholder,
                min_values=min_values,
                max_values=max_values,
                channel_types=options,
                disabled=disabled,
                custom_id=custom_id,
            )
        elif _type is discord.ComponentType.mentionable_select or type is MentionableSelect:
            self.dropdown: discord.ui.Select = MentionableSelect(
                placeholder=placeholder,
                min_values=min_values,
                max_values=max_values,
                disabled=disabled,
                custom_id=custom_id,
            )
        elif _type is discord.ComponentType.role_select or type is RoleSelect:
            self.dropdown: discord.ui.Select = RoleSelect(
                placeholder=placeholder,
                min_values=min_values,
                max_values=max_values,
                disabled=disabled,
                custom_id=custom_id,
            )
        elif _type is discord.ComponentType.user_select or _type is UserSelect:
            self.dropdown: discord.ui.Select = UserSelect(
                placeholder=placeholder,
                min_values=min_values,
                max_values=max_values,
                disabled=disabled,
                custom_id=custom_id,
            )
        else:
            if inspect.isclass(_type):
                self.dropdown: discord.ui.Select = _type(
                    placeholder=placeholder,
                    min_values=min_values,
                    max_values=max_values,
                    disabled=disabled,
                    custom_id=custom_id,
                )
            else:
                self.dropdown = _type
            setattr(self.dropdown, "callback", partial(_Select.callback, self.dropdown))
        self.add_item(self.dropdown)
        self.done: asyncio.Event = asyncio.Event()

    def to_dict_cogsutils(
        self, for_Config: typing.Optional[bool] = False
    ) -> typing.Dict[str, typing.Any]:
        dropdown_dict_instance = self.dropdown_dict_instance
        if for_Config:
            dropdown_dict_instance["members"] = None
            dropdown_dict_instance["check"] = None
            dropdown_dict_instance["function"] = None
        return dropdown_dict_instance

    @classmethod
    def from_dict_cogsutils(
        cls, dropdown_dict_instance: typing.Dict
    ) -> typing.Any:  # typing_extensions.Self
        if "function_args" in dropdown_dict_instance:
            dropdown_dict_instance["function_kwargs"] = dropdown_dict_instance["function_args"]
            del dropdown_dict_instance["function_args"]
        return cls(**dropdown_dict_instance)

    async def on_timeout(self) -> None:
        self.done.set()
        self.stop()

    async def on_error(self, interaction: discord.Interaction, error: Exception, item) -> None:
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        await interaction.followup.send("Sorry. An error has occurred.", ephemeral=True)
        discord.ui.view._log.error(
            "Ignoring exception in view %r for item %r.", self, item, exc_info=error
        )

    async def wait_result(
        self,
    ) -> typing.Tuple[discord.Interaction, typing.List[str], typing.Optional[typing.Any]]:
        self.done = asyncio.Event()
        await self.done.wait()
        interaction, options, function_result = self.get_result()
        if interaction is None:
            raise TimeoutError()
        self.interaction_result, self.options_result, self.function_result = None, None, None
        return interaction, options, function_result

    def get_result(
        self,
    ) -> typing.Tuple[discord.Interaction, typing.List[str], typing.Optional[typing.Any]]:
        return self.interaction_result, self.options_result, self.function_result

    async def callback(self, interaction: discord.Interaction) -> None:
        if self.members is not None and interaction.user.id not in self.members:
            await interaction.response.send_message(
                "You are not allowed to use this interaction.", ephemeral=True
            )
            return False
        if self.check is not None and not self.check(interaction):
            await interaction.response.send_message(
                "You are not allowed to use this interaction.", ephemeral=True
            )
            return False
        self.interaction_result = interaction
        self.options_result = self.dropdown.values
        if self.function is not None:
            self.function_result = await self.function(
                self, interaction, self.dropdown.values, **self.function_kwargs
            )
        self.done.set()
        if not self.infinity:
            self.stop()


class _Select:
    async def callback(self, interaction: discord.Interaction) -> None:
        if hasattr(self.view, "callback"):
            await self.view.callback(interaction)
        else:
            super().callback(interaction)


class Select(_Select, discord.ui.Select):
    def __init__(
        self,
        placeholder: typing.Optional[str] = "Choose an option.",
        min_values: typing.Optional[int] = 1,
        max_values: typing.Optional[int] = 1,
        *,
        options: typing.Optional[typing.List] = None,
        disabled: typing.Optional[bool] = False,
        custom_id: typing.Optional[str] = f"CogsUtils_{generate_key(length=10)}",
        row: typing.Optional[int] = None,
    ) -> None:
        if options is None:
            options = []
        super().__init__(
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            options=options,
            disabled=disabled,
            row=row,
        )


class ChannelSelect(_Select, discord.ui.ChannelSelect):
    def __init__(
        self,
        placeholder: typing.Optional[str] = "Choose a channel.",
        min_values: typing.Optional[int] = 1,
        max_values: typing.Optional[int] = 1,
        *,
        channel_types: typing.Optional[discord.ChannelType] = None,
        disabled: typing.Optional[bool] = False,
        custom_id: typing.Optional[str] = f"CogsUtils_{generate_key(length=10)}",
        row: typing.Optional[int] = None,
    ) -> None:
        if channel_types is None:
            channel_types = []
        super().__init__(
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            channel_types=channel_types,
            disabled=disabled,
            row=row,
        )


class MentionableSelect(_Select, discord.ui.MentionableSelect):
    def __init__(
        self,
        placeholder: typing.Optional[str] = "Choose an option.",
        min_values: typing.Optional[int] = 1,
        max_values: typing.Optional[int] = 1,
        *,
        disabled: typing.Optional[bool] = False,
        custom_id: typing.Optional[str] = f"CogsUtils_{generate_key(length=10)}",
        row: typing.Optional[int] = None,
    ) -> None:
        super().__init__(
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            disabled=disabled,
            row=row,
        )


class RoleSelect(_Select, discord.ui.RoleSelect):
    def __init__(
        self,
        placeholder: typing.Optional[str] = "Choose a role.",
        min_values: typing.Optional[int] = 1,
        max_values: typing.Optional[int] = 1,
        *,
        disabled: typing.Optional[bool] = False,
        custom_id: typing.Optional[str] = f"CogsUtils_{generate_key(length=10)}",
        row: typing.Optional[int] = None,
    ) -> None:
        super().__init__(
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            disabled=disabled,
            row=row,
        )


class UserSelect(_Select, discord.ui.UserSelect):
    def __init__(
        self,
        placeholder: typing.Optional[str] = "Choose a user.",
        min_values: typing.Optional[int] = 1,
        max_values: typing.Optional[int] = 1,
        *,
        disabled: typing.Optional[bool] = False,
        custom_id: typing.Optional[str] = f"CogsUtils_{generate_key(length=10)}",
        row: typing.Optional[int] = None,
    ) -> None:
        super().__init__(
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            disabled=disabled,
            row=row,
        )


class Modal(discord.ui.Modal):
    """Create Modal easily."""

    def __init__(
        self,
        title: typing.Optional[str] = "Form",
        timeout: typing.Optional[float] = None,
        inputs: typing.Optional[typing.List] = None,
        members: typing.Optional[typing.Iterable[typing.Union[discord.Member, int]]] = None,
        check: typing.Optional[typing.Callable] = None,
        function: typing.Optional[typing.Callable] = None,
        function_kwargs: typing.Optional[typing.Dict[str, typing.Any]] = None,
        custom_id: typing.Optional[str] = f"CogsUtils_{generate_key(length=10)}",
    ) -> None:
        """label: str, style: TextStyle, custom_id: str, placeholder: Optional[str], default: Optional[str], required: bool, min_length: Optional[int], max_length: Optional[int], row: Optional[int]"""
        if inputs is None:
            inputs = [{}]
        if function_kwargs is None:
            function_kwargs = {}
        for input_dict in inputs:
            if "custom_id" not in input_dict:
                input_dict["custom_id"] = f"CogsUtils_{generate_key(length=10)}"
        self.modal_dict_instance = {
            "title": title,
            "timeout": timeout,
            "inputs": [i.copy() for i in inputs],
            "members": members,
            "check": check,
            "function": function,
            "function_kwargs": function_kwargs,
            "custom_id": custom_id,
        }
        super().__init__(title=title, timeout=timeout, custom_id=custom_id)
        self.title: str = title
        self.interaction_result: typing.Optional[discord.Interaction] = None
        self.inputs_result: typing.Optional[typing.List[discord.ui.TextInput]] = None
        self.function_result: typing.Optional[typing.Any] = None
        self.members: typing.Optional[typing.List[int]] = (
            members if members is None else [getattr(member, "id", member) for member in members]
        )
        self.check: typing.Optional[typing.Callable] = check
        self.function: typing.Optional[typing.Callable] = function
        self.function_kwargs: typing.Optional[typing.Dict[str, typing.Any]] = function_kwargs
        self.inputs: typing.List[discord.ui.TextInput] = []
        self.inputs_dict: typing.List[typing.Dict[str, typing.Any]] = []
        for input_dict in inputs:
            if "label" not in input_dict:
                input_dict["label"] = "Test"
            if "style" in input_dict and isinstance(input_dict["style"], int):
                input_dict["style"] = discord.TextStyle(input_dict["style"])
            _input = discord.ui.TextInput(**input_dict)
            self.add_item(_input)
            self.inputs.append(_input)
            self.inputs_dict.append(input_dict)
        self.done: asyncio.Event = asyncio.Event()

    def to_dict_cogsutils(
        self, for_Config: typing.Optional[bool] = False
    ) -> typing.Dict[str, typing.Any]:
        modal_dict_instance = self.modal_dict_instance
        if for_Config:
            modal_dict_instance["members"] = None
            modal_dict_instance["check"] = None
            modal_dict_instance["function"] = None
        return modal_dict_instance

    @classmethod
    def from_dict_cogsutils(
        cls, modal_dict_instance: typing.Dict
    ) -> typing.Any:  # typing_extensions.Self
        if "function_args" in modal_dict_instance:
            modal_dict_instance["function_kwargs"] = modal_dict_instance["function_args"]
            del modal_dict_instance["function_args"]
        return cls(**modal_dict_instance)

    async def on_submit(self, interaction: discord.Interaction) -> None:
        if self.members is not None and interaction.user.id not in self.members:
            await interaction.response.send_message(
                "You are not allowed to use this interaction.", ephemeral=True
            )
            return False
        if self.check is not None and not self.check(interaction):
            await interaction.response.send_message(
                "You are not allowed to use this interaction.", ephemeral=True
            )
            return False
        self.interaction_result = interaction
        self.inputs_result = self.inputs
        if self.function is not None:
            self.function_result = await self.function(
                self, self.interaction_result, self.inputs_result, **self.function_kwargs
            )
        self.done.set()
        self.stop()

    async def on_timeout(self) -> None:
        self.done.set()
        self.stop()

    async def wait_result(
        self,
    ) -> typing.Tuple[
        discord.Interaction, typing.List[discord.ui.TextInput], typing.Optional[typing.Any]
    ]:
        self.done = asyncio.Event()
        await self.done.wait()
        interaction, inputs_result, function_result = self.get_result()
        if interaction is None:
            raise TimeoutError()
        self.interaction_result, self.inputs_result, self.function_result = None, None, None
        return interaction, inputs_result, function_result

    def get_result(
        self,
    ) -> typing.Tuple[
        discord.Interaction, typing.List[discord.ui.TextInput], typing.Optional[typing.Any]
    ]:
        return self.interaction_result, self.inputs_result, self.function_result


import setuptools  # isort:skip
from pathlib import Path
import re

# Get the base directory
base_dir = Path(__file__).resolve().parent

# Read the long description
try:
    long_description = (base_dir / "README.md").read_text()
except FileNotFoundError:
    long_description = "Utils for StarBot."

# Read the version from __version__.py
try:
    content = (base_dir / "Scarlet_Utils" / "__version__.py").read_text()
    __version__ = re.compile(r"__version__\s*=\s*['\"](?P<version>\d+\.\d+\.\d+)['\"]").search(content).groupdict()["version"]
except FileNotFoundError:
    __version__ = "0.0.1"  # Default version if __version__.py is missing

# Package setup
setuptools.setup(
    name="Scarlet_Utils",
    version=__version__,
    author="Scarlet Jean",
    author_email="r79767525@gmail.com",
    description="Utils for StarBot.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deathamongstlife/Star-Discord-Utils",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.9.1",
    install_requires=[
        "sentry_sdk",
        "colorama",
    ],
    extras_require={
        "dev": ["pytest", "black"],
        "docs": ["sphinx"],
    },
)